import mock

from datetime import datetime, timedelta

from plone.app.event.base import default_timezone

from seantis.dir.events.tests import BrowserTestCase
from seantis.dir.events.dates import default_now


class CommonBrowserTests(BrowserTestCase):

    def create_fetch_entry(self, **kw):
        def_start = datetime.today().replace(second=0) + timedelta(days=10)
        def_end = def_start + timedelta(hours=1)
        defaults = {
            # from IDirectoryItem
            # description not used
            'title': '',

            # from IEventsDirectoryItem
            # submitter, submitter_email not used
            'short_description': '',
            'long_description': '',
            'image': '',
            'attachment_1': '',
            'attachment_2': '',
            'locality': '',
            'street': '',
            'housenumber': '',
            'zipcode': '',
            'town': '',
            'location_url': '',
            'event_url': '',
            'organizer': '',
            'contact_name': '',
            'contact_email': '',
            'contact_phone': '',
            'prices': '',
            'registration': '',

            # from IExternalEvent
            # source not used (filled in by ExternalEventImporter)
            'source_id': 'id-1',

            # From IEventBasic
            'start': def_start,
            'end': def_end,
            'whole_day': False,
            'timezone': default_timezone(),

            # From IEventRecurrence
            'recurrence': '',

            # additional attributes used to control import
            'fetch_id': 'fetch-1',
            'last_update': default_now().replace(microsecond=0),
            'latitude': '',
            'longitude': '',
            'cat1': set(),
            'cat2': set(),
        }

        for attr in defaults:
            if not attr in kw:
                kw[attr] = defaults[attr]

        return kw

    def addGuidleSource(self, title='Guidle Source', limit=25):
        browser = self.admin_browser

        # Add form
        browser.open('/veranstaltungen/++add++seantis.dir.events.sourceguidle')
        browser.widget('title').value = title
        browser.widget('url').value = 'http://localhost:8888/'
        browser.getControl('Enabled').selected = True
        browser.widget('limit').value = str(limit)

        # Save
        browser.getControl('Save').click()

        # Check if saved
        browser.open('/veranstaltungen/')
        self.assertTrue(title in browser.contents)

    @mock.patch('seantis.dir.events.sources.guidle.EventsSourceGuidle.fetch')
    def test_browser_import_guidle(self, fetch):
        anom = self.new_browser()
        admin = self.admin_browser

        self.addGuidleSource(title='GS1')
        self.addGuidleSource(title='GS2')
        self.addGuidleSource(title='GS3', limit=1)

        # Import events (two for each source)
        fetch.return_value = [
            self.create_fetch_entry(title='event1', source_id='event1'),
            self.create_fetch_entry(title='event2', source_id='event2')
        ]
        anom.open('/veranstaltungen/fetch?run=1&no_shuffle=True')

        # Anonymous users can't see import filters and sources
        anom.open('/veranstaltungen/')
        self.assertTrue('GS1' not in anom.contents)
        self.assertTrue('GS2' not in anom.contents)
        self.assertTrue('GS3' not in anom.contents)

        # Admins can see filters and sources
        admin.open('/veranstaltungen/')
        self.assertTrue('GS1' in admin.contents)
        self.assertTrue('GS2' in admin.contents)
        self.assertTrue('GS3' in admin.contents)
        self.assertTrue('/gs1' in admin.contents)
        self.assertTrue('/gs2' in admin.contents)
        self.assertTrue('/gs3' in admin.contents)

        # Filters
        admin.open('/veranstaltungen?source=gs1')
        self.assertTrue('/event1' in admin.contents)
        self.assertTrue('/event1-1' not in admin.contents)
        self.assertTrue('/event1-2' not in admin.contents)
        admin.open('/veranstaltungen?source=gs2')
        self.assertTrue('/event2-1' in admin.contents)
        self.assertTrue('/event2-2' not in admin.contents)

        # Limit
        admin.open('/veranstaltungen?source=gs3')
        self.assertTrue('/event1-2' in admin.contents)
        self.assertTrue('/event2-2' not in admin.contents)

        # Imported Events can be hidden
        admin.open('/veranstaltungen?source=')
        self.assertTrue('Hide' in admin.contents)
        self.assertTrue('Archive' not in admin.contents)

        # Hide an event
        admin.open('/veranstaltungen/event1-2/do-action?action=hide')
        admin.open('/veranstaltungen')
        self.assertTrue('/event1-2' not in admin.contents)
        admin.open('/veranstaltungen?state=hidden')
        self.assertTrue('Publish' in admin.contents)
        self.assertTrue('/event1-2' in admin.contents)

        # Re-publish it
        admin.open('/veranstaltungen/event1-2/do-action?action=publish')
        admin.open('/veranstaltungen?state=published')
        self.assertTrue('/event1-2' in admin.contents)

        # Imported events may not be edited
        admin.open('/veranstaltungen/event2/edit')
        self.assertTrue('/gs1' in admin.contents)
        self.assertTrue('Title' not in admin.contents)
        self.assertTrue('Description' not in admin.contents)
        self.assertTrue('Save Event' not in admin.contents)

        # Imported events are not exported
        anom.open('/veranstaltungen?type=json')
        self.assertTrue('event1' not in anom.contents)

        # ... unless we request it!
        anom.open('/veranstaltungen?type=json&imported=1')
        self.assertTrue('event1' in anom.contents)

    def addSourceContentRule(self, title='rule', source=''):
        browser = self.admin_browser

        # Add rule
        browser.open('/+rule/plone.ContentRule')
        browser.getControl('Title').value = title
        browser.getControl('Object added to this container').selected = True
        browser.getControl('Save').click()

        # Add rule condition
        self.assertTrue('Event Import Source' in browser.contents)
        browser.getControl('Event Import Source').selected = True
        browser.getControl(name='form.button.AddCondition').click()
        self.assertTrue('The source id to check for.' in browser.contents)
        browser.getControl(name='form.source').value = source
        browser.getControl('Save').click()

        # Add rule action
        message = 'message_' + title + '_' + source
        browser.getControl('Notify user').selected = True
        browser.getControl(name='form.button.AddAction').click()
        browser.getControl(name='form.message').value = message
        browser.getControl('Save').click()

        # Assign rule
        browser.getControl('Apply rule on the whole site').click()

        # Check if saved
        browser.open('/@@rules-controlpanel')
        self.assertTrue(title in browser.contents)

    @mock.patch('seantis.dir.events.sources.guidle.EventsSourceGuidle.fetch')
    def test_browser_import_content_rule(self, fetch):
        browser = self.admin_browser

        # Add first guidle source
        self.addGuidleSource(title='gs1')

        # Add rules
        self.addSourceContentRule(title='all')
        self.addSourceContentRule(title='gs2', source='gs2')

        # Import events (one for each source)
        fetch.return_value = [
            self.create_fetch_entry(title='event'),
        ]
        browser.open('/veranstaltungen/fetch?run=1&force=true&no_shuffle=True')
        browser.open('/veranstaltungen')
        self.assertTrue('message_all_' in browser.contents)
        self.assertTrue('message_gs2_' not in browser.contents)

        # Add second guidle source
        self.addGuidleSource(title='gs2')
        browser.open('/veranstaltungen/fetch?run=1&force=true&no_shuffle=True')
        browser.open('/veranstaltungen')
        self.assertTrue('message_all_' in browser.contents)
        self.assertTrue('message_gs2_' in browser.contents)



from Products.CMFCore.WorkflowCore import WorkflowException
from zope.interface import alsoProvides

from seantis.dir.events.interfaces import IExternalEvent
from seantis.dir.events.tests import IntegrationTestCase


class TestWorkflow(IntegrationTestCase):

    def setUp(self):
        super(TestWorkflow, self).setUp()
        # login to gain the right to create events
        self.login_testuser()

    def tearDown(self):
        super(TestWorkflow, self).tearDown()
        self.logout()

    def test_events_workflow(self):

        # ensure that the initial state is preview
        event = self.create_event()
        self.assertEqual(event.state, 'preview')

        # go through the states
        event.submit()
        self.assertEqual(event.state, 'submitted')

        event.deny()
        self.assertEqual(event.state, 'archived')

        event = self.create_event()
        self.assertEqual(event.state, 'preview')

        event.submit()
        self.assertEqual(event.state, 'submitted')

        event.publish()
        self.assertEqual(event.state, 'published')

        event.archive()
        self.assertEqual(event.state, 'archived')

        event.publish()
        self.assertEqual(event.state, 'published')

        alsoProvides(event, IExternalEvent)

        event.hide()
        self.assertEqual(event.state, 'hidden')

        event.publish()
        self.assertEqual(event.state, 'published')

        # try some impossible changes
        event = self.create_event()
        action = lambda a: event.do_action(a)

        self.assertRaises(WorkflowException, action, 'archive')
        self.assertRaises(WorkflowException, action, 'deny')
        self.assertRaises(WorkflowException, action, 'hide')
        self.assertRaises(WorkflowException, action, 'publish')

        event.submit()

        self.assertRaises(WorkflowException, action, 'archive')
        self.assertRaises(WorkflowException, action, 'hide')
        self.assertRaises(WorkflowException, action, 'submit')

        event.publish()

        self.assertRaises(WorkflowException, action, 'deny')
        self.assertRaises(WorkflowException, action, 'hide')
        self.assertRaises(WorkflowException, action, 'publish')
        self.assertRaises(WorkflowException, action, 'submit')

        event.archive()

        self.assertRaises(WorkflowException, action, 'archive')
        self.assertRaises(WorkflowException, action, 'deny')
        self.assertRaises(WorkflowException, action, 'hide')
        self.assertRaises(WorkflowException, action, 'submit')

        event.publish()
        alsoProvides(event, IExternalEvent)
        event.hide()

        self.assertRaises(WorkflowException, action, 'archive')
        self.assertRaises(WorkflowException, action, 'deny')
        self.assertRaises(WorkflowException, action, 'hide')
        self.assertRaises(WorkflowException, action, 'submit')

    def test_access_guard(self):

        # normal events
        event = self.create_event()
        allow = ['submit', 'publish', 'deny', 'archive']
        deny = ['hide']
        for action in allow:
            self.assertTrue(event.allow_action(action))
        for action in deny:
            self.assertFalse(event.allow_action(action))

        # imported events
        alsoProvides(event, IExternalEvent)
        allow = ['hide', 'publish']
        deny = ['submit', 'deny', 'archive']
        for action in allow:
            self.assertTrue(event.allow_action(action))
        for action in deny:
            self.assertFalse(event.allow_action(action))

        # use a custom guard to ensure that the guards are actually
        # set in the definition xml file
        event = self.create_event()
        action = lambda a: event.do_action(a)

        event.allow_action = lambda a: False

        self.assertRaises(WorkflowException, action, 'submit')

        event.allow_action = lambda a: True

        action('submit')



import pytz

from itertools import groupby
from collections import OrderedDict
from datetime import datetime, timedelta

from zope.proxy import ProxyBase
from urllib import urlencode

from plone.event.recurrence import recurrence_sequence_ical
from plone.event.utils import utcoffset_normalize, DSTADJUST

from seantis.dir.events import dates
from seantis.dir.events.dates import overlaps


class Occurrence(ProxyBase):

    """ plone.app.event creates occurrences using adapters, which seems
    a bit wasteful to me. It also doesn't play too well with my interfaces,
    overridding either my title/description attributes ore not
    showing my methods.

    This pure proxy has none of those problems.

    ========================================================================
    THERE IS ONE CAVEAT. In templates "context/start" will yield the wrapped
    item's start. "python: context.start" on the other hand will yield the
    occurence's start. Same with other properties/methods
    ========================================================================
    """

    __slots__ = (
        '_wrapped', '_start', '_end', '_unsplit_start', '_unsplit_end'
    )

    def __init__(self, item, start, end, unsplit_start=None, unsplit_end=None):
        self._wrapped = item
        self._start = start
        self._end = end
        self._unsplit_start = unsplit_start
        self._unsplit_end = unsplit_end

    def __new__(cls, item, *args, **kwargs):
        return ProxyBase.__new__(cls, item)

    def get_object(self):
        """ Wraps the underlying getObject of the brain, ensuring that the
        result is still an occurrence with the same characteristics.

        """
        if hasattr(self._wrapped, 'getObject'):
            return Occurrence(
                self._wrapped.getObject(),
                self._start,
                self._end,
                self._unsplit_start,
                self._unsplit_end
            )
        else:
            return self

    @property
    def tz(self):
        return pytz.timezone(self.timezone)

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    @property
    def local_start(self):
        return self.tz.normalize(self.start)

    @property
    def local_end(self):
        return self.tz.normalize(self.end)

    @property
    def real_start(self):
        return self._wrapped.start

    @property
    def real_end(self):
        return self._wrapped.end

    @property
    def real_local_start(self):
        return self.tz.normalize(self._wrapped.start)

    @property
    def real_local_end(self):
        return self.tz.normalize(self._wrapped.end)

    @property
    def unsplit_start(self):
        return self._unsplit_start

    @property
    def unsplit_end(self):
        return self._unsplit_end

    def url(self):
        """ Adds the date of the occurrence to the result of absolute_url. This
        allows to distinguish between occurrences.

        There would be two other options, but they are less ideal:

        1. Add the whole date-time string - Results in a less than pretty url.
        2. Add an occurrence count - The underlying item of the url might
        change without any chance to tell the user.

        """
        base = self._wrapped.absolute_url()

        # don't add anything if the wrapped item doesn't have recurrence
        if not self._wrapped.recurrence:
            return base

        base += '?' in base and '&' or '?'
        return base + urlencode({'date': self._start.strftime('%Y-%m-%d')})

    def human_date(self, request):
        return dates.human_date(self.local_start, request)

    def human_daterange(self, request):
        # occurrences split to days all get the same date string
        if all((self.unsplit_start, self.unsplit_end)):
            start = self.unsplit_start
            end = self.unsplit_end
        else:
            start = self.start
            end = self.end

        start = self.tz.normalize(start)
        end = self.tz.normalize(end)

        return dates.human_daterange(start, end, request)


def grouped_occurrences(occurrences, request):
    """ Returns the given occurrences grouped by human_date. """

    def groupkey(item):
        date = item.human_date(request)
        return date

    groups = groupby(sorted(occurrences, key=lambda o: o.start), groupkey)

    # Zope Page Templates don't know how to handle generators :-|
    result = OrderedDict()

    for group in groups:
        result[group[0]] = [i for i in group[1]]

    return result


def pick_occurrence(item, start):
    """ Returns the occurrence at startdate. This suffices because two
    occurrences of the same item are never on the same day.

    """

    min_date = datetime(
        start.year, start.month, start.day,
        tzinfo=item.local_start.tzinfo
    )
    max_date = min_date + timedelta(days=1, microseconds=-1)

    found = occurrences(item, min_date, max_date)
    assert len(found) in (0, 1), "Multiple occurences on the same day?"

    return found and found[0] or None


def split_days(occurrence):
    """ Iterates through a list of occurrences and splits the occurrences
    which span more than 24 hours into sub-occurrences. The idea is to
    have events that last multiple days display on each day separately.

    """
    original_start = occurrence.local_start
    original_end = occurrence.local_end

    days = dates.split_days_count(original_start, original_end)

    if not days:
        yield occurrence
    else:
        for day in xrange(0, days + 1):
            first_day = day == 0
            last_day = day == days

            if first_day:
                start = original_start
            else:
                start = datetime(
                    original_start.year,
                    original_start.month,
                    original_start.day,
                    tzinfo=original_start.tzinfo
                ) + timedelta(days=day)

            if last_day:
                end = original_end
            else:
                end = datetime(
                    original_start.year,
                    original_start.month,
                    original_start.day,
                    tzinfo=original_start.tzinfo
                ) + timedelta(days=day + 1, microseconds=-1)

            start, end = map(dates.to_utc, (start, end))

            # if the given occurrence is already a proxy, don't double-wrap it
            if isinstance(occurrence, Occurrence):
                yield Occurrence(
                    occurrence._wrapped,
                    start, end, original_start, original_end
                )
            else:
                yield Occurrence(
                    occurrence,
                    start, end, original_start, original_end
                )


def occurrences_over_limit(rule, start, limit):

    assert rule and start and limit

    for i, o in enumerate(recurrence_sequence_ical(start=start, recrule=rule)):
        if i + 1 > limit:
            return True

    return False


def occurrences(item, min_date, max_date):
    """ Returns the occurrences for item between min and max date.
    Will return a list with a single item if the given item has no recurrence.

    """

    if not isinstance(item.start, datetime):
        item_start = dates.to_utc(datetime.utcfromtimestamp(item.start))
    else:
        item_start = item.start

    if not isinstance(item.end, datetime):
        item_end = dates.to_utc(datetime.utcfromtimestamp(item.end))
    else:
        item_end = item.end

    if not item.recurrence:

        if not overlaps(min_date, max_date, item_start, item_end):
            return []
        else:
            return [Occurrence(item, item_start, item_end)]

    tz = pytz.timezone(item.timezone)
    local_start = tz.normalize(item_start)

    _occurrences = recurrence_sequence_ical(
        start=local_start,
        recrule=item.recurrence,
        from_=min_date,
        until=max_date
    )

    result = []
    duration = item_end - item_start

    for start in _occurrences:
        start = utcoffset_normalize(start, dstmode=DSTADJUST)
        result.append(Occurrence(item, start, start + duration))

    return result


def has_future_occurrences(item, reference_date):

    if not item.recurrence:
        return reference_date <= item.start

    futures = recurrence_sequence_ical(
        item.start, recrule=item.recurrence, from_=reference_date
    )

    try:
        return bool(futures.next())
    except StopIteration:
        return False



import functools
import time
import json
import pytz

from datetime import datetime

from Products.CMFCore.utils import getToolByName

from seantis.dir.base.interfaces import IDirectoryCatalog

from zope.component import getMultiAdapter, getAdapter
from zope.component.hooks import getSite
from zope import i18n

from plone.namedfile import NamedFile

from collective.geo.geographer.interfaces import IGeoreferenced


def get_catalog(directory):
    return getAdapter(directory, IDirectoryCatalog)


def get_current_language(request):
    """ Returns the current language """
    portal_state = getMultiAdapter(
        (getSite(), request), name=u'plone_portal_state'
    )
    return portal_state.language()


def translate(request, text, domain='seantis.dir.events'):
    lang = get_current_language(request)
    return i18n.translate(text, target_language=lang, domain=domain)


def render_ical_response(request, context, calendar):
    name = '%s.ics' % context.getId()
    request.RESPONSE.setHeader('Content-Type', 'text/calendar; charset=UTF-8')
    request.RESPONSE.setHeader('Content-Disposition',
                               'attachment; filename="%s"' % name)
    return calendar.to_ical()


def render_json_response(request, items, compact):
    request.response.setHeader("Content-Type", "application/json")
    request.response.setHeader("Access-Control-Allow-Origin", "*")  # CORS

    utc = pytz.timezone('utc')
    duplicates = set()

    result = []
    for idx, item in enumerate(items):
        if compact:
            if item.id in duplicates:
                continue
            duplicates.add(item.id)

        event = {}

        start = item.start
        end = item.end

        if not start.tzinfo:
            start = utc.localize(start)
        if not end.tzinfo:
            end = utc.localize(end)

        updated = item.modification_date.asdatetime().replace(microsecond=0)
        event['last_update'] = updated.isoformat()
        event['id'] = item.id
        event['url'] = item.url()
        event['title'] = item.title
        event['short_description'] = item.short_description
        event['long_description'] = item.long_description
        event['cat1'] = item.cat1
        event['cat2'] = item.cat2
        event['start'] = utc.normalize(start).isoformat()
        event['end'] = utc.normalize(end).isoformat()
        event['recurrence'] = item.recurrence if compact else ''
        event['whole_day'] = item.whole_day
        event['timezone'] = item.timezone
        event['locality'] = item.locality
        event['street'] = item.street
        event['housenumber'] = item.housenumber
        event['zipcode'] = item.zipcode
        event['town'] = item.town
        event['location_url'] = item.location_url
        event['organizer'] = item.organizer
        event['contact_name'] = item.contact_name
        event['contact_email'] = item.contact_email
        event['contact_phone'] = item.contact_phone
        event['prices'] = item.prices
        event['event_url'] = item.event_url
        event['registration'] = item.registration
        event['submitter'] = item.submitter
        event['submitter_email'] = item.submitter_email

        event['images'] = []
        if isinstance(item.image, NamedFile):
            image = {}
            image['name'] = item.image.filename
            image['url'] = item.absolute_url() + '/@@images/image'
            event['images'].append(image)

        event['attachements'] = []
        if isinstance(item.attachment_1, NamedFile):
            attachement = {}
            attachement['name'] = item.attachment_1.filename
            attachement['url'] = item.absolute_url()
            attachement['url'] += '/@@download/attachment_1'
            event['attachements'].append(attachement)
        if isinstance(item.attachment_2, NamedFile):
            attachement = {}
            attachement['name'] = item.attachment_2.filename
            attachement['url'] = item.absolute_url()
            attachement['url'] += '/@@download/attachment_2'
            event['attachements'].append(attachement)

        try:
            geo = IGeoreferenced(item)
            event['longitude'] = geo.coordinates[0]
            event['latitude'] = geo.coordinates[1]
        except TypeError:
            event['longitude'] = None
            event['latitude'] = None

        result.append(event)

    return json.dumps(result)


def workflow_tool():
    return getToolByName(getSite(), "portal_workflow")


def verify_wkt(data):
    try:
        from shapely import wkt
        geom = wkt.loads(data)
    except ImportError:
        from pygeoif.geometry import from_wkt
        geom = from_wkt(data)
    return geom


def webcal(fn):
    """ Replaces the http in a function that returns an url with webcal.
    Does nothing if https is used, as webcals seems to be supported badly.
    """

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        url = fn(*args, **kwargs)

        if not url.startswith('https') and url.startswith('http'):
            url = url.replace('http', 'webcal')

        return url

    return wrapper



from datetime import datetime, timedelta
from seantis.dir.events.tests import BrowserTestCase


class CommonBrowserTests(BrowserTestCase):

    def test_json_export(self):
        # Add events
        self.addEvent(title='test1', description='desc1')
        self.addEvent(title='test2', description='desc2',
                      cat1='Category1_2', cat2='Category2_2',
                      recurrence='RRULE:FREQ=DAILY;COUNT=2')
        first_date = datetime.today().replace(
            hour=14, minute=0, second=0, microsecond=0
        )
        second_date = first_date + timedelta(days=1)

        browser = self.new_browser()

        # Export all
        browser.open('/veranstaltungen?type=json')
        self.assertEqual(browser.headers['Content-type'], 'application/json')
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' in browser.contents)
        self.assertTrue('desc1' in browser.contents)
        self.assertTrue('desc2' in browser.contents)
        self.assertTrue('Category1' in browser.contents)
        self.assertTrue('Category1_2' in browser.contents)
        self.assertTrue('Category2' in browser.contents)
        self.assertTrue('Category2_2' in browser.contents)
        self.assertTrue(first_date.isoformat() in browser.contents)
        self.assertTrue(second_date.isoformat() in browser.contents)

        # Export compact
        browser.open('/veranstaltungen?type=json&compact=1')
        self.assertTrue('RRULE:FREQ=DAILY;COUNT=2' in browser.contents)
        self.assertTrue(first_date.isoformat() in browser.contents)
        self.assertTrue(second_date.isoformat() not in browser.contents)

        # Export by using filter (&filter=&cat1=&cat2=)
        browser.open('/veranstaltungen?type=json&filter=true&cat1=Category1')
        self.assertEqual(browser.headers['Content-type'], 'application/json')
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' not in browser.contents)

        browser.open('/veranstaltungen?type=json&filter=true&cat1=Category1_2')
        self.assertEqual(browser.headers['Content-type'], 'application/json')
        self.assertTrue('test1' not in browser.contents)
        self.assertTrue('test2' in browser.contents)

        browser.open('/veranstaltungen?type=json&filter=true&cat2=Category2_2')
        self.assertEqual(browser.headers['Content-type'], 'application/json')
        self.assertTrue('test1' not in browser.contents)
        self.assertTrue('test2' in browser.contents)

        browser.open('/veranstaltungen?type=json&filter=true&cat1=C1&cat2=C2')
        self.assertEqual(browser.headers['Content-type'], 'application/json')
        self.assertTrue('test1' not in browser.contents)
        self.assertTrue('test2' not in browser.contents)

        browser.open('/veranstaltungen?type=json&cat1=Category1')
        self.assertEqual(browser.headers['Content-type'], 'application/json')
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' in browser.contents)

        # Export by using search (search=&searchtext=)
        browser.open('/veranstaltungen?type=json&search=true&searchtext=test1')
        self.assertEqual(browser.headers['Content-type'], 'application/json')
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' not in browser.contents)

        browser.open('/veranstaltungen?type=json&search=true&searchtext=test2')
        self.assertEqual(browser.headers['Content-type'], 'application/json')
        self.assertTrue('test1' not in browser.contents)
        self.assertTrue('test2' in browser.contents)

        browser.open('/veranstaltungen?type=json&search=true&searchtext=test7')
        self.assertEqual(browser.headers['Content-type'], 'application/json')
        self.assertTrue('test1' not in browser.contents)
        self.assertTrue('test2' not in browser.contents)

        browser.open('/veranstaltungen?type=json&searchtext=test1')
        self.assertEqual(browser.headers['Content-type'], 'application/json')
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' in browser.contents)

        # Export only first event
        browser.open('/veranstaltungen?type=json&max=1')
        self.assertEqual(browser.headers['Content-type'], 'application/json')
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' not in browser.contents)

    def test_issue_17(self):
        self.addEvent(title='test1')
        self.addEvent(title='test2', date=datetime.today() - timedelta(days=2),
                      check_submitted=False, do_publish=False)
        self.addEvent(title='test3', date=datetime.today() - timedelta(days=4),
                      check_submitted=False, do_publish=False)
        self.admin_browser.open(
            '/veranstaltungen/test3/do-action?action=publish')

        # Past events (submitted or published) should not be exported
        browser = self.new_browser()
        browser.open('/veranstaltungen?type=json')
        self.assertEqual(browser.headers['Content-type'], 'application/json')
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' not in browser.contents)
        self.assertTrue('test3' not in browser.contents)

    def test_ical_export(self):
        # Add events
        self.addEvent(title='test1', description='desc1')
        self.addEvent(title='test2', description='desc2',
                      cat1='Category1_2', cat2='Category2_2')

        browser = self.new_browser()

        # Export all
        browser.open('/veranstaltungen?type=ical')
        self.assertTrue('text/calendar' in browser.headers['Content-type'])
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' in browser.contents)

        # Export by using filter (&filter=&cat1=&cat2=)
        browser.open('/veranstaltungen?type=ical&filter=true&cat1=Category1')
        self.assertTrue('text/calendar' in browser.headers['Content-type'])
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' not in browser.contents)

        browser.open('/veranstaltungen?type=ical&filter=true&cat1=Category1_2')
        self.assertTrue('text/calendar' in browser.headers['Content-type'])
        self.assertTrue('test1' not in browser.contents)
        self.assertTrue('test2' in browser.contents)

        browser.open('/veranstaltungen?type=ical&filter=true&cat2=Category2_2')
        self.assertTrue('text/calendar' in browser.headers['Content-type'])
        self.assertTrue('test1' not in browser.contents)
        self.assertTrue('test2' in browser.contents)

        browser.open('/veranstaltungen?type=ical&filter=true&cat1=C1&cat2=C2')
        self.assertTrue('text/calendar' in browser.headers['Content-type'])
        self.assertTrue('test1' not in browser.contents)
        self.assertTrue('test2' not in browser.contents)

        browser.open('/veranstaltungen?type=ical&cat1=Category1')
        self.assertTrue('text/calendar' in browser.headers['Content-type'])
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' in browser.contents)

        # Export by using search (search=&searchtext=)
        browser.open('/veranstaltungen?type=ical&search=true&searchtext=test1')
        self.assertTrue('text/calendar' in browser.headers['Content-type'])
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' not in browser.contents)

        browser.open('/veranstaltungen?type=ical&search=true&searchtext=test2')
        self.assertTrue('text/calendar' in browser.headers['Content-type'])
        self.assertTrue('test1' not in browser.contents)
        self.assertTrue('test2' in browser.contents)

        browser.open('/veranstaltungen?type=ical&search=true&searchtext=test7')
        self.assertTrue('text/calendar' in browser.headers['Content-type'])
        self.assertTrue('test1' not in browser.contents)
        self.assertTrue('test2' not in browser.contents)

        browser.open('/veranstaltungen?type=ical&searchtext=test1')
        self.assertTrue('text/calendar' in browser.headers['Content-type'])
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' in browser.contents)

        # Export via item view
        browser.open('/veranstaltungen/test1?type=ical')
        self.assertTrue('text/calendar' in browser.headers['Content-type'])
        self.assertTrue('test1' in browser.contents)

        browser.open('/veranstaltungen/test2?type=ical')
        self.assertTrue('text/calendar' in browser.headers['Content-type'])
        self.assertTrue('test2' in browser.contents)



from mock import Mock

from seantis.dir.events.tests import IntegrationTestCase
from seantis.dir.events import maintenance


class TestMaintenance(IntegrationTestCase):

    def setUp(self):
        super(TestMaintenance, self).setUp()
        self.old_logger = maintenance.log
        maintenance.log = Mock()

    def tearDown(self):
        super(TestMaintenance, self).tearDown()
        maintenance.log = self.old_logger

    def test_register_clock_server(self):
        result = maintenance.register('method', 60 * 60)
        self.assertEquals(result.method, 'method')
        self.assertTrue('method' in maintenance._clockservers)

        maintenance.clear_clockservers()
        self.assertTrue('method' not in maintenance._clockservers)

    def test_clock_logger(self):
        logger = maintenance.ClockLogger('method')

        logger.log('GET http://localhost:888/method HTTP/1.1 200')
        self.assertEquals(0, len(maintenance.log.method_calls))

        logger.log('GET http://localhost:888/method HTTP/1.1 500')
        self.assertTrue('call.warn' in str(maintenance.log.method_calls))
        self.assertTrue('500' in str(maintenance.log.method_calls))

        logger.log('GET http://localhost:888/method HTTP/1.1')
        self.assertTrue('call.error' in str(maintenance.log.method_calls))

        logger.log('')
        self.assertEquals(3, len(maintenance.log.method_calls))

        logger.log('abcd')
        self.assertEquals(4, len(maintenance.log.method_calls))



import re
import mock

from datetime import datetime, timedelta
from seantis.dir.events import dates
from seantis.dir.events.tests import BrowserTestCase


class CommonBrowserTests(BrowserTestCase):

    def test_workflow(self):

        # anonymous browser
        fourchan = self.new_browser()
        fourchan.open('/veranstaltungen/@@submit')

        self.assertTrue('Send us your events' in fourchan.contents)

        # create some event
        def create_event():
            fourchan.getControl(name='form.widgets.title').value = 'Party'
            fourchan.getControl(
                name='form.widgets.short_description'
            ).value = 'Some Party'

            fourchan.set_date('form.widgets.submission_date', datetime.now())
            fourchan.getControl(
                name='form.widgets.submission_start_time'
            ).value = '08:00 AM'
            fourchan.getControl(
                name='form.widgets.submission_end_time'
            ).value = '12:00 PM'

            fourchan.getControl('Category1').selected = True
            fourchan.getControl('Category2').selected = True

            fourchan.getControl('Continue').click()
            self.assertTrue('preview' in fourchan.url)
            fourchan.getControl('Continue').click()
            self.assertTrue('finish' in fourchan.url)

            fourchan.getControl(
                name='form.widgets.submitter'
            ).value = 'John Doe'
            fourchan.getControl(
                name='form.widgets.submitter_email'
            ).value = 'john.doe@example.com'

            fourchan.getControl('Submit').click()

        create_event()

        # the event is now invisible to the anonymous user in the directory
        fourchan.open('/veranstaltungen?state=submitted')
        self.assertFalse('Some Party' in fourchan.contents)

        # it is however visible to the admin
        browser = self.admin_browser
        browser.open('/veranstaltungen?state=submitted')

        self.assertTrue('Some Party' in browser.contents)

        browser = self.admin_browser
        browser.open('/veranstaltungen?state=submitted')

        self.assertTrue('Some Party' in browser.contents)

        # unless the admin filters it out
        browser.open('/veranstaltungen?state=published')

        self.assertFalse('Some Party' in browser.contents)

        # who now has the chance to either deny or publish the event
        browser.open('/veranstaltungen?state=submitted')
        self.assertTrue('Publish' in browser.contents)
        self.assertTrue('Deny Publication' in browser.contents)
        self.assertTrue('Submitted' in browser.contents)

        # let's deny
        browser.getLink('Deny Publication').click()

        # the event should now be invisible to both admin and anonymous
        browser.open('/veranstaltungen')
        self.assertFalse('Some Party' in browser.contents)

        fourchan.open(browser.url)
        self.assertFalse('Some Party' in browser.contents)

        # let's repeat, but publish this time
        fourchan.open('/veranstaltungen/@@submit')
        create_event()

        browser.open('/veranstaltungen?state=submitted')
        browser.getLink('Publish', index=1).click()

        # this should've led to a state change
        browser.open('/veranstaltungen?state=published')
        self.assertTrue('Some Party' in browser.contents)
        self.assertTrue('Archive' in browser.contents)

        fourchan.open(browser.url)
        self.assertTrue('Some Party' in fourchan.contents)
        self.assertFalse('Archive' in fourchan.contents)

        # archiving the event should hide it again
        browser.getLink('Archive').click()
        browser.open('/veranstaltungen')

        self.assertFalse('Some Party' in browser.contents)

        fourchan.open(browser.url)
        self.assertFalse('SomeParty' in browser.contents)

    def test_preview(self):

        # anonymous browser
        fourchan = self.new_browser()
        fourchan.open('/veranstaltungen/@@submit')

        self.assertTrue('Send us your events' in fourchan.contents)

        # create a recurring event
        fourchan.widget('title').value = 'Recurring'
        fourchan.widget('short_description').value = 'Every Day'
        fourchan.widget('locality').value = 'at home'

        fourchan.set_date('submission_date', datetime.now())

        fourchan.widget('submission_start_time').value = '08:00 AM'
        fourchan.widget('submission_end_time').value = '04:00 PM'

        fourchan.getControl('Category1').selected = True
        fourchan.getControl('Category2').selected = True

        fourchan.widget(
            'submission_recurrence'
        ).value = 'RRULE:FREQ=DAILY;COUNT=7'

        fourchan.getControl('Continue').click()
        self.assertTrue('preview' in fourchan.url)

        # expect all fields to be shown and the recurrence resulting in
        # a number of events in the list

        self.assertTrue('Recurring' in fourchan.contents)
        self.assertTrue('Every Day' in fourchan.contents)
        self.assertTrue('at home' in fourchan.contents)
        self.assertEqual(fourchan.contents.count('"eventgroup"'), 7)

        # update the recurrence and check back
        fourchan.getControl('Adjust').click()

        fourchan.widget(
            'submission_recurrence'
        ).value = 'RRULE:FREQ=DAILY;COUNT=365'

        fourchan.getControl('Continue').click()

        self.assertEqual(fourchan.contents.count('"eventgroup"'), 365)

        # remove the recurrence, ensuring that one event remains
        fourchan.getControl('Adjust').click()

        fourchan.widget('submission_recurrence').value = ''
        fourchan.getControl('Continue').click()

        self.assertEqual(fourchan.contents.count('"eventgroup"'), 1)

        # ensure that no more than 365 occurrences may be entered
        fourchan.getControl('Adjust').click()
        fourchan.widget(
            'submission_recurrence'
        ).value = 'RRULE:FREQ=DAILY;COUNT=366'

        fourchan.getControl('Continue').click()

        self.assertTrue(
            'You may not add more than 365 occurences' in fourchan.contents
        )

        # regression test for an issue where occurrences in the future
        # were not counted correctly
        fourchan.set_date('submission_date', datetime(2021, 1, 1, 0, 0))
        fourchan.widget('submission_start_time').value = '10:00 AM'
        fourchan.widget('submission_end_time').value = '11:00 AM'

        fourchan.widget(
            'submission_recurrence'
        ).value = 'RRULE:FREQ=DAILY;UNTIL=20211231T000000'

        fourchan.getControl('Continue').click()  # ok

        self.assertFalse(
            'You may not add more than 365 occurences' in fourchan.contents
        )

        fourchan.getControl('Adjust').click()

        fourchan.widget(
            'submission_recurrence'
        ).value = 'RRULE:FREQ=DAILY;UNTIL=20220101T000000'

        fourchan.getControl('Continue').click()  # not okay

        self.assertTrue(
            'You may not add more than 365 occurences' in fourchan.contents
        )

    def test_event_submission(self):

        browser = self.admin_browser

        # get a browser for anonymous
        fourchan = self.new_browser()
        fourchan.open('/veranstaltungen')

        self.assertTrue('Veranstaltungen' in browser.contents)

        # get to the submit form
        fourchan.getLink('Submit Your Event').click()

        self.assertTrue('Send us your events' in fourchan.contents)

        fourchan.getControl(name='form.widgets.title').value = 'Stammtisch'
        fourchan.getControl(
            name='form.widgets.short_description'
        ).value = 'Socializing Yo'

        fourchan.set_date('form.widgets.submission_date', datetime.now())
        fourchan.getControl(
            name='form.widgets.submission_start_time'
        ).value = '08:00 AM'
        fourchan.getControl(
            name='form.widgets.submission_end_time'
        ).value = '12:00 AM'

        fourchan.getControl('Category1').selected = True
        fourchan.getControl('Category2').selected = True

        fourchan.getControl('Continue').click()

        # previewing an event should send us to the preview view
        self.assertTrue('preview' in fourchan.url)

        # a token should have been added to the url
        self.assertTrue('token=' in fourchan.url)

        # the preview should contain the entered information
        self.assertTrue('Socializing Yo' in fourchan.contents)

        # if the user tries to submit another event while this one is still
        # in preview, the existing event is loaded
        # (the form is turned into an edit form)
        oldurl = fourchan.url

        fourchan.open('/veranstaltungen/@@submit')
        self.assertEqual(
            fourchan.getControl(name='form.widgets.title').value,
            'Stammtisch'
        )
        self.assertEqual(
            fourchan.getControl(name='form.widgets.short_description').value,
            'Socializing Yo'
        )

        fourchan.open(oldurl)

        # there's a change-event button which submits a GET request to
        # the submit form using the token in the request
        fourchan.getControl('Adjust').click()
        self.assertTrue('submit?token=' in fourchan.url)
        self.assertFalse(fourchan.url.endswith('?token='))

        # we should be able to change some things
        # and come back to the url to find those changes

        fourchan.getControl(
            name='form.widgets.short_description'
        ).value = 'Serious Business'
        fourchan.getControl('Continue').click()

        self.assertTrue('Serious Business' in fourchan.contents)
        self.assertTrue('preview' in fourchan.url)

        # at the same time this event in preview is invisble in the list
        # even for administrators
        browser.open('/veranstaltungen')
        self.assertTrue('Veranstaltungen' in browser.contents)
        self.assertFalse('Serious Business' in browser.contents)

        # other anonymous users may not access the view or the preview
        google_robot = self.new_browser()
        google_robot.assert_notfound('/veranstaltungen/stammtisch')
        google_robot.assert_notfound(
            '/veranstaltungen/stammtisch/preview'
        )

        # not event the admin at this point (not sure about that one yet)
        browser.assert_notfound('/veranstaltungen/stammtisch')
        browser.assert_notfound(
            '/veranstaltungen/stammtisch/preview'
        )

        # if the user decides to cancel the event before submitting it, he
        # loses the right to access the event (will be cleaned up by cronjob)
        fourchan.getControl('Cancel').click()

        fourchan.assert_notfound('/veranstaltungen/stammtisch')
        fourchan.assert_notfound(
            '/veranstaltungen/stammtisch/preview'
        )
        fourchan.assert_notfound(
            '/veranstaltungen/stammtisch/edit-event'
        )

        # since we cancelled we must now create a new event to
        # test the submission process
        new = self.new_browser()
        new.open('/veranstaltungen/@@submit')

        self.assertEqual(new.getControl(name='form.widgets.title').value, '')
        self.assertEqual(
            new.getControl(name='form.widgets.short_description').value, ''
        )

        new.set_date('form.widgets.submission_date', datetime.now())
        new.getControl(
            name='form.widgets.submission_start_time'
        ).value = '08:00 AM'
        new.getControl(
            name='form.widgets.submission_end_time'
        ).value = '12:00 AM'

        new.getControl('Category1').selected = True
        new.getControl('Category2').selected = True

        new.getControl(name='form.widgets.title').value = "Submitted Event"
        new.getControl(name='form.widgets.short_description').value = "YOLO"

        new.getControl('Continue').click()

        # at this point the event is invisble to the admin
        browser.open('/veranstaltungen?state=submitted')
        self.assertFalse('YOLO' in browser.contents)

        # until the anonymous user submits the event
        new.getControl('Continue').click()

        new.getControl('Submit').click()

        # at this point we 'forgot' to fill in the submitter info so we have at
        # it again and fix that
        new.getControl(name='form.widgets.submitter').value = 'John Doe'
        new.getControl(
            name='form.widgets.submitter_email'
        ).value = 'john.doe@example.com'

        new.getControl('Submit').click()

        browser.open('/veranstaltungen?state=submitted')
        self.assertTrue('YOLO' in browser.contents)

        # the user may no longer access the event at this point, though
        # it is no longer an inexistant resource
        new.assert_unauthorized('/veranstaltungen/submitted-event')

        # the admin should be able to see the submitter's name and email
        browser.open('/veranstaltungen/submitted-event')
        self.assertTrue('John Doe' in browser.contents)
        self.assertTrue('john.doe@example.com' in browser.contents)

        # once we publish it and open in another browser this information is
        # hidden from the public eye
        url = browser.url
        browser.open((
            '/veranstaltungen/submitted-event'
            '/content_status_modify?workflow_action=publish'
        ))

        public = self.new_browser()
        public.open(url)

        self.assertTrue('YOLO' in public.contents)
        self.assertFalse('John Doe' in public.contents)
        self.assertFalse('john.doe@example.com' in public.contents)

    def test_simplified_recurrence_submission(self):
        browser = self.admin_browser
        browser.open('/veranstaltungen/++add++seantis.dir.events.item')

        browser.widget('title').value = 'Testtitle'
        browser.widget('short_description').value = 'Testdescription'

        browser.getControl('Category1').selected = True
        browser.getControl('Category2').selected = True

        browser.widget('submission_date_type').value = ['range']

        start = datetime.today()
        end = start + timedelta(days=2)

        browser.set_date('submission_range_start_date', start)
        browser.set_date('submission_range_end_date', end)
        browser.widget('submission_range_start_time').value = '2:00 PM'
        browser.widget('submission_range_end_time').value = '4:00 PM'

        browser.getControl('Continue').click()

        self.assertTrue('3 Occurrences' in browser.contents)

        browser.getControl('Adjust').click()

        browser.set_date('submission_range_end_date', end - timedelta(days=1))

        browser.getControl('Continue').click()

        self.assertTrue('2 Occurrences' in browser.contents)

        browser.getControl('Adjust').click()

        browser.widget('submission_date_type').value = ['date']
        browser.set_date('submission_date', start)
        browser.widget('submission_start_time').value = '2:00 PM'
        browser.widget('submission_end_time').value = '4:00 PM'
        browser.widget('submission_recurrence').value = ''

        browser.getControl('Continue').click()

        self.assertTrue('No Occurrences' in browser.contents)

        browser.getControl('Adjust').click()
        browser.widget('submission_date_type').value = ['range']
        browser.set_date('submission_range_start_date', start)
        browser.set_date('submission_range_end_date', start)

        browser.getControl('Continue').click()

        self.assertTrue('No Occurrences' in browser.contents)

    def test_simplified_recurrence_submission_days(self):
        browser = self.admin_browser
        browser.open('/veranstaltungen/++add++seantis.dir.events.item')

        browser.widget('title').value = 'Testtitle'
        browser.widget('short_description').value = 'Testdescription'

        browser.getControl('Category1').selected = True
        browser.getControl('Category2').selected = True

        browser.widget('submission_date_type').value = ['range']

        # get a monday to test with
        start = datetime.now()
        while start.weekday() != 0:
            start += timedelta(days=1)

        end = start + timedelta(days=6)  # sunday

        browser.set_date('submission_range_start_date', start)
        browser.set_date('submission_range_end_date', end)
        browser.widget('submission_range_start_time').value = '2:00 PM'
        browser.widget('submission_range_end_time').value = '4:00 PM'
        browser.widget('submission_days:list').value = ['MO', 'SU']

        browser.getControl('Continue').click()
        self.assertTrue('2 Occurrences' in browser.contents)

        browser.getControl('Adjust').click()
        browser.widget('submission_days:list').value = ['MO', 'WE', 'SU']

        browser.getControl('Continue').click()
        self.assertTrue('3 Occurrences' in browser.contents)

        browser.getControl('Adjust').click()
        browser.widget('submission_days:list').value = []

        browser.getControl('Continue').click()
        self.assertTrue('No Occurrences' in browser.contents)

    def test_default_forms(self):

        # admins use the submit / preview forms for adding / editing
        # as well so we don't have to support two different form types
        # the following code tests that
        browser = self.admin_browser

        browser.open('/veranstaltungen/++add++seantis.dir.events.item')
        self.assertTrue('Send us your events' in browser.contents)

        browser.getControl(name='form.widgets.title').value = 'Add Test'
        browser.getControl(
            name='form.widgets.short_description'
        ).value = 'Add Test Description'

        browser.set_date('form.widgets.submission_date', datetime.now())
        browser.getControl(
            name='form.widgets.submission_start_time'
        ).value = '08:00 AM'
        browser.getControl(
            name='form.widgets.submission_end_time'
        ).value = '04:00 PM'

        browser.getControl('Category1').selected = True
        browser.getControl('Category2').selected = True

        browser.getControl('Continue').click()
        self.assertTrue('preview' in browser.url)

        self.assertTrue('Add Test Description' in browser.contents)

        browser.getControl('Continue').click()
        self.assertTrue('finish' in browser.url)

        browser.getControl('Submitter Name').value = 'Submitter'
        browser.getControl('Submitter Email').value = 'submit@example.com'

        browser.getControl('Submit').click()

        # show the submitted events
        browser.getLink('Submitted').click()

        self.assertTrue('Add Test Description' in browser.contents)
        self.assertTrue('Veranstaltungen' in browser.contents)

        browser.open('/veranstaltungen/add-test/edit')
        self.assertTrue('Send us your events' in browser.contents)

        browser.getControl(
            name='form.widgets.short_description'
        ).value = 'Changed Test Description'
        browser.getControl('Save Event').click()

        self.assertTrue('Changed Test Description' in browser.contents)
        self.assertFalse('preview' in browser.url)

    def test_recurrence(self):
        browser = self.admin_browser

        browser.open(
            '/veranstaltungen/++add++seantis.dir.events.item'
        )
        self.assertTrue('Send us your events' in browser.contents)

        browser.getControl(name='form.widgets.title').value = 'Recurring'
        browser.getControl(
            name='form.widgets.short_description'
        ).value = 'Add Test Description'
        browser.getControl(
            name='form.widgets.submission_recurrence'
        ).value = 'RRULE:FREQ=DAILY;COUNT=7'

        browser.set_date('form.widgets.submission_date', datetime.now())
        browser.getControl(
            name='form.widgets.submission_start_time'
        ).value = '08:00 AM'
        browser.getControl(
            name='form.widgets.submission_end_time'
        ).value = '04:00 PM'

        browser.getControl('Category1').selected = True
        browser.getControl('Category2').selected = True

        browser.getControl('Continue').click()
        browser.getControl('Continue').click()

        browser.getControl('Submitter Name').value = 'Submitter'
        browser.getControl('Submitter Email').value = 'submit@example.com'

        browser.getControl('Submit').click()

        # make sure to get all events
        browser.getLink('This and Next Year').click()

        # with the state submitted
        browser.getLink('Submitted').click()

        # take the last occurrence
        first_url = browser.getLink('Recurring', index=0).url
        link = browser.getLink('Recurring', index=6)
        year, month, day = map(
            int, link.url[len(link.url) - 10:].split('-')
        )

        # and ensure that the date is correct in the detail view
        link.click()

        self.assertFalse('Today' in browser.contents)
        self.assertTrue('%02i.%02i' % (day, month) in browser.contents)

        browser.open(first_url)
        self.assertTrue('Today' in browser.contents)

        # replace the date in the url with an invalid date -> 404
        first_url = re.sub('\d{4}-\d{2}-\d{2}', r'2000-01-01', first_url)
        browser.assert_notfound(first_url)

    def test_wrong_dates(self):
        browser = self.admin_browser

        # Missing start time
        browser.open('/veranstaltungen/@@submit')
        browser.widget('title').value = 'Title'
        browser.widget('short_description').value = 'Description'
        browser.getControl('Category1').selected = True
        browser.getControl('Category2').selected = True
        browser.set_date('form.widgets.submission_date',
                         datetime.now() + timedelta(days=2))
        browser.getControl(
            name='form.widgets.submission_end_time'
        ).value = '12:00 PM'
        browser.getControl('Continue').click()
        self.assertTrue('Missing start time' in browser.contents)

        # Missing end time
        browser.open('/veranstaltungen/@@submit')
        browser.widget('title').value = 'Title'
        browser.widget('short_description').value = 'Description'
        browser.getControl('Category1').selected = True
        browser.getControl('Category2').selected = True
        browser.set_date('form.widgets.submission_date',
                         datetime.now() + timedelta(days=2))
        browser.getControl(
            name='form.widgets.submission_start_time'
        ).value = '08:00 AM'
        browser.getControl('Continue').click()
        self.assertTrue('Missing end time' in browser.contents)

    def test_coordinates(self):
        browser = self.admin_browser

        # Invalid coordinates
        browser.open('/veranstaltungen/@@submit')
        browser.widget('title').value = 'event with coordinates'
        browser.widget('short_description').value = 'Short'
        browser.getControl('Category1').selected = True
        browser.getControl('Category2').selected = True
        browser.set_date('form.widgets.submission_date',
                         datetime.now() + timedelta(days=2))
        browser.getControl('All day').selected = True
        browser.getControl('Category1').selected = True
        browser.getControl('Category2').selected = True
        browser.getControl(
            name="form.widgets.wkt"
        ).value = "PINT (8.3156129 47.05033479999997)"
        click = lambda: browser.getControl('Continue').click()
        self.assertRaises(Exception, click)

        # Valid coordinates
        browser.open('/veranstaltungen/@@submit')
        browser.widget('title').value = 'event with coordinates'
        browser.widget('short_description').value = 'Short'
        browser.getControl('Category1').selected = True
        browser.getControl('Category2').selected = True
        browser.set_date('form.widgets.submission_date',
                         datetime.now() + timedelta(days=2))
        browser.getControl('All day').selected = True
        browser.getControl('Category1').selected = True
        browser.getControl('Category2').selected = True
        browser.getControl(
            name="form.widgets.wkt"
        ).value = "POINT (8.3156129 47.05033479999997)"
        browser.getControl('Continue').click()
        browser.getControl('Continue').click()
        browser.getControl('Submitter Name').value = 'Submitter'
        browser.getControl('Submitter Email').value = 'submit@example.com'
        browser.getControl('Submit').click()

        browser.open('/veranstaltungen/event-with-coordinates/edit')
        self.assertTrue(
            "POINT (8.3156129 47.05033479999997)" in browser.contents
        )

    def test_terms(self):
        browser = self.admin_browser

        def enable_terms(enable):
            new = self.new_browser()
            new.login_admin()

            new.open('/veranstaltungen/edit')
            new.getControl(
                name="form.widgets.terms"
            ).value = enable and 'verily, though agreeth' or ''
            new.getControl('Save').click()

        browser.open(
            '/veranstaltungen/++add++seantis.dir.events.item'
        )
        self.assertTrue('Send us your events' in browser.contents)

        browser.getControl(name='form.widgets.title').value = 'Test'
        browser.getControl(
            name='form.widgets.short_description'
        ).value = 'Test'

        browser.set_date('form.widgets.submission_date', datetime.now())
        browser.getControl(
            name='form.widgets.submission_start_time'
        ).value = '08:00 AM'
        browser.getControl(
            name='form.widgets.submission_end_time'
        ).value = '04:00 PM'

        browser.getControl('Category1').selected = True
        browser.getControl('Category2').selected = True

        browser.getControl('Continue').click()
        browser.getControl('Continue').click()

        self.assertFalse('Terms and Conditions' in browser.contents)

        enable_terms(True)

        browser.reload()
        self.assertTrue('Terms and Conditions' in browser.contents)

        browser.getLink('Terms and Conditions').click()
        self.assertTrue('verily' in browser.contents)

        browser.goBack()

        # if not agreed upon, the submission is denied
        browser.getControl('Submitter Name').value = 'Submitter'
        browser.getControl('Submitter Email').value = 'submit@example.com'
        self.assertTrue('Terms and Conditions' in browser.contents)

        browser.getControl('Submit').click()
        self.assertTrue('finish' in browser.url)
        self.assertTrue('Terms and Conditions' in browser.contents)

        browser.getControl(name='form.widgets.agreed:list').value = 'selected'
        browser.getControl('Submit').click()

        self.assertFalse('finish' in browser.url)
        self.assertFalse('Terms and Conditions' in browser.contents)
        self.assertTrue('Event submitted' in browser.contents)

    @mock.patch('seantis.dir.events.dates.default_timezone')
    @mock.patch('seantis.dir.events.dates.default_now')
    def test_whole_day_regression(self, default_now, default_timezone):
        # recurring whole day events would lead to errors because the
        # daterange applied in the catalog did not include 00:00 one day
        # after creating the event

        # the error only happens if the timezone of the event and the
        # server differs

        default_timezone.return_value = 'Europe/Vienna'
        default_now.return_value = dates.as_timezone(
            datetime.now(), 'Europe/Vienna'
        )

        browser = self.admin_browser

        # create the event in europe/vienna timezone
        browser.open(
            '/veranstaltungen/++add++seantis.dir.events.item'
        )

        browser.widget('title').value = 'Title'
        browser.widget('short_description').value = 'Short'

        browser.getControl('Category1').selected = True
        browser.getControl('Category2').selected = True

        browser.widget('submission_date_type').value = ['range']

        start = datetime.now()
        end = datetime.now() + timedelta(days=2)

        browser.set_date('submission_range_start_date', start)
        browser.set_date('submission_range_end_date', end)
        browser.getControl('All day').selected = True

        browser.getControl('Continue').click()
        browser.getControl('Continue').click()

        browser.widget('submitter').value = 'test'
        browser.widget('submitter_email').value = 'test@example.com'

        browser.getControl('Submit').click()

        # view the event the next day in UTC
        default_timezone.return_value = 'UTC'
        default_now.return_value = dates.as_timezone(
            datetime.now() + timedelta(days=1), 'Europe/Vienna'
        )

        # this used to throw an error
        browser.open('/veranstaltungen?state=submitted')

    def test_filter_and_search(self):
        # Add events
        self.addEvent(title='test1', description='desc1')
        self.addEvent(title='test2', description='desc2',
                      cat1='Category1_2', cat2='Category2_2')

        browser = self.new_browser()

        # No categories, no filter
        browser.open('/veranstaltungen')
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' in browser.contents)

        # Filter (&filter=&cat1=&cat2=)
        browser.open('/veranstaltungen?filter=true&cat1=Category1')
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' not in browser.contents)

        browser.open('/veranstaltungen?filter=true&cat1=Category1_2')
        self.assertTrue('test1' not in browser.contents)
        self.assertTrue('test2' in browser.contents)

        browser.open('/veranstaltungen?filter=true&cat2=Category2_2')
        self.assertTrue('test1' not in browser.contents)
        self.assertTrue('test2' in browser.contents)

        browser.open('/veranstaltungen?filter=true&cat1=Category2_2')
        self.assertTrue('test1' not in browser.contents)
        self.assertTrue('test2' not in browser.contents)

        # Search (search=&searchtext=)
        browser.open('/veranstaltungen?search=true&searchtext=test1')
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' not in browser.contents)

        browser.open('/veranstaltungen?search=true&searchtext=test2')
        self.assertTrue('test1' not in browser.contents)
        self.assertTrue('test2' in browser.contents)

        browser.open('/veranstaltungen?search=true&searchtext=token')
        self.assertTrue('test1' not in browser.contents)
        self.assertTrue('test2' not in browser.contents)

    def test_eventindex_view(self):
        # Test unauthorized access
        anonymous = self.new_browser()
        anonymous.assert_unauthorized('/veranstaltungen/eventindex')

        browser = self.admin_browser

        # Test eventindex
        browser.open('/veranstaltungen/eventindex')
        self.assertTrue('text/plain' in browser.headers['Content-type'])
        self.assertTrue('test1' not in browser.contents)
        self.assertTrue('test2' not in browser.contents)

        self.addEvent(title='test1')
        browser.open('/veranstaltungen/eventindex')
        self.assertTrue('text/plain' in browser.headers['Content-type'])
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' not in browser.contents)

        self.addEvent(title='test2')
        browser.open('/veranstaltungen/eventindex')
        self.assertTrue('text/plain' in browser.headers['Content-type'])
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' in browser.contents)

        # reindex
        browser.open('/veranstaltungen/eventindex?reindex')
        self.assertTrue('text/plain' in browser.headers['Content-type'])
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' in browser.contents)

        # rebuild
        browser.open('/veranstaltungen/eventindex?rebuild')
        self.assertTrue('text/plain' in browser.headers['Content-type'])
        self.assertTrue('test1' in browser.contents)
        self.assertTrue('test2' in browser.contents)

    def test_browser_add_today_whole_day(self):
        browser = self.admin_browser

        self.addEvent(title='whole-day-event-today',
                      whole_day=True,
                      check_submitted=False, do_publish=False)

        browser.open('/veranstaltungen?state=submitted')

        # The event should be visible (it wasn't in an earlier version)
        self.assertTrue('whole-day-event-today' in browser.contents)
        self.assertTrue('Submitted (1)' in browser.contents)

        # .. .also if published
        browser.getLink('Publish', index=1).click()
        browser.open('/veranstaltungen?state=published')
        self.assertTrue('whole-day-event-today' in browser.contents)

    def test_date_range_selection(self):
        browser = self.admin_browser

        def open_range(date1, date2):
            browser.open(
                'veranstaltungen?range=custom&from=%s&to=%s' % (
                    '%04i-%02i-%02i' % (date1.year, date1.month, date1.day),
                    '%04i-%02i-%02i' % (date2.year, date2.month, date2.day)
                )
            )

        dates = [datetime.today() + timedelta(days=d) for d in [8, 15, 28, 45]]
        for date in dates:
            self.addEvent(title=str(date), date=date)

        open_range(datetime.today(), dates[2])
        self.assertTrue(str(dates[0]) in browser.contents)
        self.assertTrue(str(dates[1]) in browser.contents)
        self.assertTrue(str(dates[2]) in browser.contents)
        self.assertFalse(str(dates[3]) in browser.contents)

        open_range(dates[1], dates[3])
        self.assertFalse(str(dates[0]) in browser.contents)
        self.assertTrue(str(dates[1]) in browser.contents)
        self.assertTrue(str(dates[2]) in browser.contents)
        self.assertTrue(str(dates[3]) in browser.contents)

        open_range(dates[3], dates[2])
        self.assertFalse(str(dates[0]) in browser.contents)
        self.assertFalse(str(dates[1]) in browser.contents)
        self.assertFalse(str(dates[2]) in browser.contents)
        self.assertTrue(str(dates[3]) in browser.contents)

    def test_cleanup_view(self):
        browser = self.admin_browser
        browser.open('veranstaltungen/cleanup?run=1')

    def test_fetch_view(self):
        browser = self.admin_browser
        browser.open('veranstaltungen/fetch')



from os import urandom
from uuid import UUID
from five import grok

from zope.component.hooks import getSite
from zope.component import getAdapter
from zExceptions import NotFound

from Products.CMFCore.utils import getToolByName

from seantis.dir.base.session import get_session, set_session
from seantis.dir.events.interfaces import ITokenAccess, IEventsDirectoryItem


class TokenAccess(grok.Adapter):

    grok.context(IEventsDirectoryItem)
    grok.provides(ITokenAccess)

    def attach_token(self, token=None):
        assert token != 'missing'

        self.context.access_token = token or UUID(bytes=urandom(16)).hex
        store_on_session(self.context)

    def has_access(self, request):

        if not hasattr(self.context, 'access_token'):
            return False

        token = current_token(request)
        if token == self.context.access_token:
            store_on_session(self.context)
            return True

        return False

    def clear_token(self):
        if hasattr(self.context, 'access_token'):
            del self.context.access_token

        remove_from_session()


def current_token(request):
    token = request.get('token', 'missing')

    if token == 'missing':
        try:
            token = retrieve_from_session()
        except AttributeError:
            pass

    return token.replace('-', '')


def store_on_session(context):
    assert context.access_token
    set_session(getSite(), 'events-access-token', context.access_token)


def retrieve_from_session():
    return get_session(getSite(), 'events-access-token') or 'missing'


def remove_from_session():
    set_session(getSite(), 'events-access-token', None)


def apply_token(context):
    token_access = getAdapter(context, ITokenAccess)
    token_access.attach_token()


def verify_token(context, request):
    if not context.state == 'preview':
        return

    token_access = getAdapter(context, ITokenAccess)

    if not token_access.has_access(request):
        raise NotFound()


def clear_token(context):
    token_access = getAdapter(context, ITokenAccess)
    token_access.clear_token()


def append_token(context, url):
    if not hasattr(context, 'access_token'):
        return url

    querychar = '?' if '?' not in url else '&'

    return url + querychar + 'token=' + context.access_token


def event_by_token(directory, token):

    if token == 'missing':
        return None

    catalog = getToolByName(directory, 'portal_catalog')
    path = '/'.join(directory.getPhysicalPath())

    results = catalog(path={'query': path, 'depth': 1},
        object_provides=IEventsDirectoryItem.__identifier__,
        review_state='preview',
    )

    for result in results:
        obj = result.getObject()

        if not hasattr(obj, 'access_token'):
            continue

        if obj.access_token == token:
            return obj

    return None



import functools

from zope.component.hooks import getSite

from AccessControl import ClassSecurityInfo, getSecurityManager
from AccessControl.SecurityManagement import newSecurityManager, setSecurityManager
from AccessControl.User import nobody
from AccessControl.User import UnrestrictedUser as BaseUnrestrictedUser

class UnrestrictedUser(BaseUnrestrictedUser):
    """Unrestricted user that still has an id.
    """
    def getId(self):
        """Return the ID of the user.
        """
        return self.getUserName()

def execute_under_special_role(portal, role, function, *args, **kwargs):
    """ Execute code under special role priviledges.

    Example how to call::

        execute_under_special_role(portal, "Manager",
            doSomeNormallyNotAllowedStuff,
            source_folder, target_folder)


    @param portal: Reference to ISiteRoot object whose access controls we are using

    @param function: Method to be called with special priviledges

    @param role: User role for the security context when calling the priviledged code; e.g. "Manager".

    @param args: Passed to the function

    @param kwargs: Passed to the function
    """

    sm = getSecurityManager()

    try:
        try:
            # Clone the current user and assign a new role.
            # Note that the username (getId()) is left in exception
            # tracebacks in the error_log,
            # so it is an important thing to store.
            tmp_user = UnrestrictedUser(
                sm.getUser().getId(), '', [role], ''
                )

            # Wrap the user in the acquisition context of the portal
            tmp_user = tmp_user.__of__(portal.acl_users)
            newSecurityManager(None, tmp_user)

            # Call the function
            return function(*args, **kwargs)

        except:
            # If special exception handlers are needed, run them here
            raise
    finally:
        # Restore the old security manager
        setSecurityManager(sm)


import pytz

from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta, MO, FR

from plone.app.event.base import default_timezone
from seantis.dir.events import utils
from seantis.dir.events import _

default_daterange = 'this_year'


def eventrange():
    """ Returns the daterange (start, end) in which events may exist. """
    now = default_now()
    this_morning = datetime(now.year, now.month, now.day)

    # to work in all timezones the lowest timezone has to be considered
    # for the morning (the last place where day x is still occurring)
    # according to http://en.wikipedia.org/wiki/List_of_UTC_time_offsets this
    # is Baker / Howland Island with an UTC Offset of -12
    return (
        this_morning - timedelta(seconds=60 * 60 * 12),
        this_morning + timedelta(days=365 * 2)
    )


def overlaps(start, end, otherstart, otherend):
    """ Returns true if the two date ranges somehow overlap. """

    if otherstart <= start and start <= otherend:
        return True

    if start <= otherstart and otherstart <= end:
        return True

    return False


def default_now():
    """ Returns the current time using the default_timezone from
    plone.app.event. Said timezone is either set globally or per user.

    """
    utcnow = to_utc(datetime.utcnow())
    return pytz.timezone(default_timezone()).normalize(utcnow)


def to_utc(date):
    """ Converts date to utc, making it timezone aware if not already. """
    return as_timezone(date, 'utc')


def delete_timezone(date):
    """ Removes the timezone from a date. """
    assert date.tzinfo
    return date.replace(tzinfo=None)


def as_timezone(date, timezone):
    """ Converts date to the given timezone, making it timezone aware if not
    already. Use this instead of date.astimezone, since said function does
    not account for daylight savings time. """
    timezone = pytz.timezone(timezone)
    if not date.tzinfo:
        date = timezone.localize(date)

    return timezone.normalize(date)


def as_rfc5545_string(datetime):
    """ Converts a datetime into the RFC5545 Datetime Form #2 as defined in
    http://tools.ietf.org/html/rfc5545#section-3.3.5

    """
    return to_utc(datetime).strftime('%Y%m%dT%H%M%SZ')


def is_whole_day(start, end):
    return all((
        start.hour == 0,
        start.minute == 0,
        start.second == 0,
        end.hour == 23,
        end.minute == 59,
        end.second == 59
    ))


def days_between(start, end):
    return [(start + timedelta(days=d)) for d in xrange((end - start).days)]


def combine_daterange(date, start_time, end_time):
    """ Combines a date, a start_time and an end_time into a start-datetime,
    end-datetime range.

    """
    if end_time >= start_time:
        return (
            datetime.combine(date, start_time),
            datetime.combine(date, end_time)
        )
    else:
        return (
            datetime.combine(date, start_time),
            datetime.combine(date + timedelta(days=1), end_time)
        )


def split_days_count(start, end):
    """ Returns 0 if the given daterange must be kept together and a number
    of splits that need to be created if not.

    If the event ends between 0:00 and 08:59 the new date is not counted as
    a new day. An event that goes through the night is not a two-day event.

    """

    days = (end.date() - start.date()).days

    if days == 0:
        return 0

    if 0 <= end.hour and end.hour <= 8:
        days -= 1

    return days


def human_date(date, request):
    now = default_now()

    if now.date() == date.date():
        return _(u'Today')

    if now.date() + timedelta(days=1) == date.date():
        return _(u'Tomorrow')

    calendar = request.locale.dates.calendars['gregorian']
    weekday = calendar.getDayNames()[date.weekday()]

    if now.year == date.year:
        return weekday + ' ' + date.strftime('%d.%m.')
    else:
        return weekday + ' ' + date.strftime('%d.%m.%Y')


def human_daterange(start, end, request):

    if is_whole_day(start, end):
        if split_days_count(start, end) < 1:
            return utils.translate(request, _(u'Whole Day'))
        else:
            if default_now().year == start.year:
                return start.strftime('%d.%m. - ') \
                    + end.strftime('%d.%m. ') \
                    + utils.translate(request, _(u'Whole Day'))
            else:
                return start.strftime('%d.%m.%Y. - ') \
                    + end.strftime('%d.%m.%Y. ') \
                    + utils.translate(request, _(u'Whole Day'))

    if split_days_count(start, end) < 1:
        return start.strftime('%H:%M - ') + end.strftime('%H:%M')
    else:
        if default_now().year == start.year:
            return start.strftime('%d.%m. %H:%M - ') \
                + end.strftime('%d.%m. %H:%M')
        else:
            return start.strftime('%d.%m.%Y. %H:%M - ') \
                + end.strftime('%d.%m.%Y. %H:%M')

methods = list()
ranges = dict()
labels = dict()
tooltips = dict()


def daterange(label, tooltip=u''):
    """ Deocrator that, applied to DateRangeInfo methods, marks them
    as category methods for further processing.

    """
    tooltip = tooltip or label

    def decorator(fn):
        global methods, ranges, labels, tooltips

        methods.append((fn.__name__, label, tooltip))
        ranges[label] = fn.__name__
        labels[fn.__name__] = label
        tooltips[fn.__name__] = tooltip

        return fn

    return decorator


def is_valid_daterange(name):
    """ Returns true if the given name is a valid DateRangeInfo method. """
    return name in ranges.values()


weekdays = dict(MO=0, TU=1, WE=2, TH=3, FR=4, SA=5, SU=6)


def this_weekend(date):
    """ Returns a daterange with the start being friday 4pm and the end
    being sunday midnight, relative to the given date.

    """
    this_morning = datetime(
        date.year, date.month, date.day, 0, 0, tzinfo=date.tzinfo
    )

    if this_morning.weekday() in (weekdays["SA"], weekdays["SU"]):
        weekend_start = this_morning + relativedelta(weekday=FR(-1))
    else:
        weekend_start = this_morning + relativedelta(weekday=FR(+1))

    weekend_end = weekend_start + relativedelta(weekday=MO(1))

    weekend_start += timedelta(hours=16)
    weekend_end += timedelta(microseconds=-1)

    return weekend_start, weekend_end


def this_month(date):
    month_start = datetime(date.year, date.month, 1, 0, 0, tzinfo=date.tzinfo)
    month_end = month_start + relativedelta(months=1, microseconds=-1)

    return month_start, month_end


def next_weekday(date, weekday):
    w = weekdays[weekday]

    if date.weekday() == w:
        return date

    return date + timedelta((w - date.weekday()) % 7)


def as_range(start, end):
    start = datetime(start.year, start.month, start.day, tzinfo=start.tzinfo)
    end = datetime(end.year, end.month, end.day, tzinfo=end.tzinfo)
    end = end + timedelta(days=1, microseconds=-1)
    return start, end


class DateRanges(object):

    def __init__(self, now=None):
        self.now = now or default_now()

    def get_now(self):
        return self._now

    def set_now(self, now):
        self._now = now

        self.this_morning = datetime(
            now.year, now.month, now.day, tzinfo=now.tzinfo
        )
        self.this_evening = self.this_morning + timedelta(
            days=1, microseconds=-1
        )

        # mornings shall start at 3 am
        self.this_morning += timedelta(hours=3)

    now = property(get_now, set_now)

    def overlaps(self, method, start, end):
        s, e = getattr(self, method)
        return overlaps(s, e, start, end)

    @property
    @daterange(_(u'Today'))
    def today(self):
        return self.this_morning, self.this_evening

    @property
    @daterange(_(u'Tomorrow'))
    def tomorrow(self):
        return (
            self.this_morning + timedelta(days=1),
            self.this_evening + timedelta(days=1)
        )

    @property
    @daterange(_(u'Day after Tomorrow'))
    def day_after_tomorrow(self):
        return (
            self.this_morning + timedelta(days=2),
            self.this_evening + timedelta(days=2)
        )

    @property
    @daterange(
        _(u'This Weekend'), _(u'From 4pm this Friday until 12pm this Sunday')
    )
    def this_weekend(self):
        # between friday at 4 and saturday midnight
        return this_weekend(self.now)

    @property
    @daterange(
        _(u'Next Weekend'), _(u'From 4pm next Friday until 12pm next Sunday')
    )
    def next_weekend(self):
        weekend_start, weekend_end = this_weekend(self.now)
        weekend_start += timedelta(days=7)
        weekend_end += timedelta(days=7)

        return weekend_start, weekend_end

    @property
    @daterange(_(u'This Week'), _(u'From today until 12pm this Sunday'))
    def this_week(self):
        # range between now and next sunday evening with
        # range contains at least two days (saturday until next sunday)
        end_of_week = next_weekday(self.this_evening + timedelta(days=2), "SU")
        return self.this_morning, end_of_week

    @property
    @daterange(
        _(u'Next Week'), _(u'From next Monday until the following Sunday')
    )
    def next_week(self):
        # range between next sunday (as in self.is_this_week)
        # and the sunday after that
        start_of_week = next_weekday(
            self.this_morning + timedelta(days=2), "SU"
        )
        start_of_week += timedelta(days=1)
        start_of_week = datetime(
            start_of_week.year, start_of_week.month, start_of_week.day, 0, 0,
            tzinfo=self.now.tzinfo
        )
        end_of_week = next_weekday(start_of_week, "SU") \
            + timedelta(days=1, microseconds=-1)

        return start_of_week, end_of_week

    @property
    @daterange(_(u'This Month'), _(u'From today until the end of the month'))
    def this_month(self):
        _, end_of_this_month = this_month(self.now)
        return self.this_morning, end_of_this_month

    @property
    @daterange(
        _(u'Next Month'), _(u'From the beginning of next month until the end')
    )
    def next_month(self):
        _, prev_end = this_month(self.now)
        month_start = prev_end + timedelta(microseconds=1)

        return this_month(month_start)

    @property
    @daterange(
        _(u'This Year'), _(u'From today until the end of this year')
    )
    def this_year(self):
        end_of_year = datetime(
            self.now.year + 1, 1, 1, tzinfo=self.now.tzinfo
        )
        end_of_year -= timedelta(microseconds=1)

        return self.this_morning, end_of_year

    @property
    @daterange(_(u'Next Year'), _(u'The whole next year'))
    def next_year(self):
        start_of_year = datetime(
            self.now.year + 1, 1, 1, tzinfo=self.now.tzinfo
        )
        end_of_year = datetime(self.now.year + 2, 1, 1, tzinfo=self.now.tzinfo)
        end_of_year -= timedelta(microseconds=1)

        return start_of_year, end_of_year

    @property
    @daterange(
        _(u'This and Next Year'), _(u'From today until the end of next year')
    )
    def this_and_next_year(self):
        return self.this_year[0], self.next_year[1]

    @property
    @daterange(_(u'From ... to ...'), _(u'Custom date range'))
    def custom(self):
        # These values are only used to as default values and are overwritten
        return self.this_month



import os

import logging
log = logging.getLogger('seantis.dir.events')

import hashlib
import transaction
import isodate
import pytz

from functools32 import lru_cache

from five import grok

from datetime import datetime, timedelta
from urllib2 import urlopen, HTTPError
from itertools import groupby
from random import shuffle
from threading import Lock
from plone.synchronize import synchronized

from zope.component.hooks import getSite

from collective.geo.geographer.interfaces import IWriteGeoreferenced
from Products.CMFCore.utils import getToolByName
from plone.namedfile import NamedFile, NamedImage
from plone.dexterity.utils import createContentInContainer
from zope.interface import alsoProvides
from zope.annotation.interfaces import IAnnotations

from seantis.dir.base import directory
from seantis.dir.base.interfaces import (
    IDirectoryCatalog,
    IDirectoryCategorized
)
from seantis.dir.events.interfaces import (
    IExternalEvent,
    IExternalEventCollector,
    IExternalEventSource,
    IEventsDirectory,
    NoImportDataException
)
from seantis.dir.events.unrestricted import execute_under_special_role


class ExternalEventImporter(object):

    def __init__(self, context):
        self.context = context

    def sources(self):
        result = []
        sources = IDirectoryCatalog(self.context).catalog(
            object_provides=IExternalEventSource.__identifier__
        )
        for source in sources:
            if source.getObject().enabled:
                result.append(source)

        return result

    @lru_cache(maxsize=50)
    def download(self, url):
        return urlopen(url, timeout=60).read()

    def disable_indexing(self):
        self.context._v_fetching = True

    def enable_indexing(self):
        self.context._v_fetching = False

    def get_last_update_time(self):
        assert self.annotation_key

        isostring = IAnnotations(self.context).get(self.annotation_key, None)

        if not isostring:
            return None
        else:
            return isodate.parse_datetime(isostring)

    def set_last_update_time(self, dt):
        assert self.annotation_key

        assert isinstance(dt, datetime)
        assert dt.tzinfo, "please use a timezone aware datetime"

        # use string to store date to ensure that the annotation
        # doesn't cause problems in the future
        annotations = IAnnotations(self.context)
        annotations[self.annotation_key] = isodate.datetime_isoformat(dt)

    def grouped_existing_events(self, source):
        events = {}
        if not source:
            return events

        catalog = getToolByName(self.context, 'portal_catalog')
        brains = catalog(
            object_provides=IExternalEvent.__identifier__, source=source
        )

        for key, items in groupby(brains, lambda brain: brain.source_id):
            if key:
                events[key] = [item.id for item in items]

        return events

    def _fetch_one(
        self, source, function, limit=None, reimport=False, source_ids=[],
        autoremove=False
    ):
        imported = []
        len_deleted = 0

        try:
            events = sorted(
                function(),
                key=lambda e: (e['last_update'], e['source_id'])
            )
        except NoImportDataException:
            log.info('no data received for %s' % source)
            return imported, len_deleted

        existing = self.grouped_existing_events(source)

        # Autoremove externally deleted events
        if autoremove:
            new_ids = [event['source_id'] for event in events]
            old_ids = existing.keys()
            delta = list(set(old_ids) - set(new_ids))
            if limit is not None:
                delta = delta[:limit]
            for source_id in delta:
                # source id's are not necessarily unique
                for event_id in existing[source_id]:
                    log.info('Deleting %s' % (event_id))
                    self.context.manage_delObjects(event_id)
                    len_deleted += 1

        if len(events) == 0:
            return imported, len_deleted

        fetch_ids = set(event['fetch_id'] for event in events)
        assert len(fetch_ids) == 1, """
            Each event needs a fetch_id which describes the id of the
            whole fetch process and is therefore the same for all events
            in a single fetch. See seantis.dir.events.source.guidle:events
        """

        self.annotation_key = hashlib.sha1(list(fetch_ids)[0]).hexdigest()
        last_update = self.get_last_update_time()
        last_update_in_run = datetime.min.replace(tzinfo=pytz.timezone('utc'))

        if not last_update:
            log.info('initial import')
            changed_offers_only = False
        elif reimport:
            log.info('reimport everything')
            changed_offers_only = False
        else:
            log.info('importing updates since {}'.format(last_update))
            changed_offers_only = True

        total = len(events) if not limit else limit

        workflowTool = getToolByName(self.context, 'portal_workflow')

        categories = dict(cat1=set(), cat2=set())

        limit_reached_id = None

        for ix, event in enumerate(events):

            if limit_reached_id and limit_reached_id != event['source_id']:
                break

            if source_ids and event['source_id'] not in source_ids:
                continue

            assert 'last_update' in event, """
                Each event needs a last_update datetime info which is used
                to determine if any changes were done. This is used for
                importing only changed events.
            """

            if last_update_in_run < event['last_update']:
                last_update_in_run = event['last_update']

            if changed_offers_only and event['source_id'] in existing:
                if event['last_update'] <= last_update:
                    continue

            # keep a set of all categories for the suggestions
            for cat in categories:
                if cat not in event:
                    event[cat] = set()
                categories[cat] |= event[cat]

            # stop at limit
            if limit and (len(imported) + 1) >= limit and not limit_reached_id:
                log.info('reached limit of %i events' % limit)
                # don't quit right away, all events of the same source_id
                # need to be imported first since they have the same
                # update_time
                limit_reached_id = event['source_id']

            # flush to disk every 500th event to keep memory usage low
            if len(imported) != 0 and len(imported) % 500 == 0:
                transaction.savepoint(True)

            log.info('importing %i/%i %s @ %s' % (
                (len(imported) + 1), total, event['title'],
                event['start'].strftime('%d.%m.%Y %H:%M')
            ))

            event['source'] = source

            # If the existing event has been hidden, we keep it hidden
            hide_event = False
            if event['source_id'] in existing:
                for event_id in existing[event['source_id']]:
                    review_state = self.context.get(event_id).review_state
                    hide_event |= review_state == 'hidden'

            # source id's are not necessarily unique as a single external
            # event might have to be represented as more than one event in
            # seantis.dir.events - therefore updating is done through
            # deleting first, adding second
            if event['source_id'] in existing:
                for event_id in existing[event['source_id']]:
                    self.context.manage_delObjects(event_id)
                del existing[event['source_id']]

            # image and attachments are downloaded
            downloads = {
                'image': NamedImage,
                'attachment_1': NamedFile,
                'attachment_2': NamedFile
            }

            def allow_download(download, url):

                if download != 'image':
                    return True

                # whitelist the images that are known to work
                # not working is *.bmp. We could convert but I'd rather
                # force people to use a sane format
                return url.lower().endswith((
                    'png', 'jpg', 'jpeg', '@@images/image'
                ))

            for download, method in downloads.items():
                url = event.get(download)
                name = download + '_name'

                if not url or not allow_download(download, url):
                    event[download] = None
                else:
                    try:
                        event[download] = method(self.download(url))
                        if name in event:
                            event[download].filename = event[name]
                    except HTTPError:
                        event[download] = None

                if name in event:
                    del event[name]

            # latitude and longitude are set through the interface
            lat, lon = event.get('latitude'), event.get('longitude')
            if 'latitude' in event:
                del event['latitude']
            if 'longitude' in event:
                del event['longitude']

            # so are categories
            cats = map(event.get, ('cat1', 'cat2'))
            del event['cat1']
            del event['cat2']

            assert 'cat3' not in event and 'cat4' not in event, """
                unsupported categories
            """

            obj = createContentInContainer(
                self.context, 'seantis.dir.events.item',
                checkConstraints=False,
                **event
            )

            # set coordinates now
            if lat and lon:
                IWriteGeoreferenced(obj).setGeoInterface(
                    'Point', map(float, (lon, lat))
                )

            # followed by the categories
            IDirectoryCategorized(obj).cat1 = list(cats[0])
            IDirectoryCategorized(obj).cat2 = list(cats[1])

            workflowTool.doActionFor(obj, 'submit')
            workflowTool.doActionFor(obj, 'publish')

            for download in downloads:
                getattr(obj, download)

            alsoProvides(obj, IExternalEvent)

            if hide_event:
                workflowTool.doActionFor(obj, 'hide')

            imported.append(obj)

        self.set_last_update_time(last_update_in_run)

        # add categories to suggestions
        for category in categories:
            key = '%s_suggestions' % category
            existing = getattr(self.context, key)
            existing = set(existing) if existing is not None else set()
            new = categories[category] | existing

            setattr(self.context, key, sorted(new))

            diff = categories[category] - existing
            if len(diff):
                log.info('added to %s %s' % (category, diff))

        log.info('committing events for %s' % source)
        transaction.commit()

        return imported, len_deleted

    def fetch_one(
        self, source, function, limit=None, reimport=False, source_ids=[],
        autoremove=False
    ):

        start = datetime.now()
        log.info('begin fetching events for %s' % source)

        self.disable_indexing()

        try:
            imported, len_deleted = self._fetch_one(
                source, function, limit, reimport, source_ids, autoremove
            )
        finally:
            self.enable_indexing()

        runtime = datetime.now() - start

        if len(imported):
            # Reindex
            log.info('reindexing commited events for %s' % source)

            for event in imported:
                event.reindexObject()

            self.context.reindexObject()

            IDirectoryCatalog(self.context).reindex()

            # Log runtime
            minutes = runtime.total_seconds() // 60
            seconds = runtime.seconds - minutes * 60
            log.info('imported %i events in %i minutes, %i seconds' % (
                len(imported), minutes, seconds
            ))

        return len(imported), len_deleted


class ExternalEventImportScheduler(object):

    _lock = Lock()

    def __init__(self):
        self.running = {}
        self.last_run = {}

    def is_importing_instance(self):
        """ Check if we are the instance which imports events.

        This is defined with an environment variable via the buildout file.
        """
        return os.getenv('seantis_events_import', False) == 'true'

    @synchronized(_lock)
    def handle_run(self, import_directory, do_stop=False):
        """Check if we can start importing or signal that we are finished.

        Note that it is possible that requests are still blocked, i.e. when
        doing the same request (e.g. '/fetch'), see ZSever/medusa/http_server.

        Note that it happend that some threads died while executing, hence
        forcing a run every 4 hours.

        Note that we need up to #sites_with_import threads
        (+ 1 if we are importing from ourselves) worst case.
        """
        return_value = False

        if import_directory not in self.running:
            self.running[import_directory] = False
        if import_directory not in self.last_run:
            self.last_run[import_directory] = datetime.now()

        if do_stop:
            self.running[import_directory] = False
            self.last_run[import_directory] = datetime.now()
        else:
            delta = datetime.now() - self.last_run[import_directory]
            if not self.running[import_directory] or delta > timedelta(hours=4):
                self.running[import_directory] = True
                self.last_run[import_directory] = datetime.now()
                return_value = True

        return return_value

    def run(self, context, reimport=False, source_ids=[], no_shuffle=False):

        len_imported = 0
        len_deleted = 0
        len_sources = 0

        if not self.is_importing_instance():
            return len_imported, len_deleted, len_sources

        import_directory = '/'.join(context.getPhysicalPath())

        if not self.handle_run(import_directory):
            log.info('already importing')
            return len_imported, len_deleted, len_sources

        log.info('begin importing sources from %s' % (import_directory))

        try:
            importer = ExternalEventImporter(context)
            sources = importer.sources()
            if not no_shuffle:
                shuffle(sources)
            for source in sources:
                len_sources += 1
                path = source.getPath()
                events, deleted = importer.fetch_one(
                    path,
                    IExternalEventCollector(source.getObject()).fetch,
                    source.getObject().limit,
                    reimport, source_ids,
                    source.getObject().autoremove)
                log.info('source %s processed' % (path))
                len_imported += events
                len_deleted += deleted
        finally:
            context.reindexObject()
            IDirectoryCatalog(context).reindex()
            log.info('importing sources from %s finished' % (import_directory))
            self.handle_run(import_directory, do_stop=True)

        return len_imported, len_deleted, len_sources


import_scheduler = ExternalEventImportScheduler()


class EventsDirectoryFetchView(grok.View, directory.DirectoryCatalogMixin):

    grok.name('fetch')
    grok.context(IEventsDirectory)

    template = None

    def render(self):

        self.request.response.setHeader("Content-type", "text/plain")

        reimport = bool(self.request.get('reimport', False))
        ids = self.request.get('source-ids', '').split(',')
        no_shuffle = bool(self.request.get('no_shuffle', False))
        do_run = self.request.get('run') == '1'

        if do_run:
            imported, deleted, sources = execute_under_special_role(
                getSite(), 'Manager',
                import_scheduler.run, self.context, reimport,
                all(ids) and ids or None, no_shuffle
            )

            return u'%i events imported from %i sources (%i deleted)' % (
                imported, sources, deleted
            )

        else:
            return u''



import os

from logging import getLogger
log = getLogger('seantis.dir.events')

from datetime import datetime, timedelta

from threading import Lock

from five import grok

from plone.synchronize import synchronized

from zope.component.hooks import getSite

from seantis.dir.base.interfaces import IDirectoryCatalog
from seantis.dir.events.dates import to_utc
from seantis.dir.events.interfaces import (
    IEventsDirectoryItem, IEventsDirectory, IExternalEvent
)
from seantis.dir.events.recurrence import has_future_occurrences
from seantis.dir.events.unrestricted import execute_under_special_role


class CleanupScheduler(object):

    _lock = Lock()

    def __init__(self):
        self.next_run = 0

    def is_cleaning_instance(self):
        """ Check if we are the instance which cleans up events.

        This is defined with an environment variable via the buildout file.
        """
        return os.getenv('seantis_events_cleanup', False) == 'true'

    def get_next_run(self, now=None):
        if now is None:
            now = datetime.now()

        # Schedule next run tomorrow at 0:30
        days = 1
        if now.hour < 1 and now.minute < 30:
            days = 0
        next_run = datetime(now.year, now.month, now.day) + timedelta(
            days=days, minutes=30)
        return next_run

    @synchronized(_lock)
    def run(self, directory, dryrun=False, force_run=False, now=None):

        if now is None:
            now = datetime.now()

        if not self.is_cleaning_instance():
            return

        if not self.next_run:
            self.next_run = self.get_next_run(now)

        if (now > self.next_run) or force_run:
            self.next_run = self.get_next_run(now)
            self.cleanup_directory(directory, dryrun=dryrun)

    def remove_stale_previews(self, directory, dryrun=False):

        catalog = IDirectoryCatalog(directory)
        query = catalog.catalog.unrestrictedSearchResults

        log.info('searching for stale previews (> 2 days old)')

        past = to_utc(datetime.utcnow() - timedelta(days=2))
        stale_previews = query(
            path={'query': directory.getPhysicalPath(), 'depth': 2},
            object_provides=IEventsDirectoryItem.__identifier__,
            review_state=('preview'),
            modified={'query': past, 'range': 'max'}
        )
        stale_previews = [p.id for p in stale_previews]

        if stale_previews:
            log.info('deleting stale previews -> %s' % str(stale_previews))
            if not dryrun:
                directory.manage_delObjects(stale_previews)
        else:
            log.info('no stale previews found')

        return stale_previews

    def archive_past_events(self, directory, dryrun=False):

        catalog = IDirectoryCatalog(directory)
        query = catalog.catalog

        log.info('archiving past events (> 2 days old)')

        # events are in the past if they have been over for two days
        # (not one, to ensure that they are really over in all timezones)
        past = to_utc(datetime.utcnow() - timedelta(days=2))
        published_events = query(
            path={'query': directory.getPhysicalPath(), 'depth': 2},
            object_provides=IEventsDirectoryItem.__identifier__,
            review_state=('published', ),
            start={'query': past, 'range': 'max'},
            end={'query': past, 'range': 'max'}
        )

        past_events = []

        for event in published_events:
            event = event.getObject()

            assert event.start < past
            assert event.end < past

            # recurring events may be in the past with one of
            # their occurrences in the future
            if not has_future_occurrences(event, past):
                # published events may be imported events
                if not IExternalEvent.providedBy(event):
                    past_events.append(event)

        ids = [p.id for p in past_events]

        if past_events:
            log.info('archiving past events -> %s' % str(ids))

            if not dryrun:
                for event in past_events:
                    event.archive()
        else:
            log.info('no past events found')

        return ids

    def remove_archived_events(self, directory, dryrun=False):

        catalog = IDirectoryCatalog(directory)
        query = catalog.catalog

        log.info('removing archived events (> 30 days old)')

        past = datetime.utcnow() - timedelta(days=30)
        archived_events = query(
            path={'query': directory.getPhysicalPath(), 'depth': 2},
            object_provides=IEventsDirectoryItem.__identifier__,
            review_state=('archived', ),
            start={'query': past, 'range': 'max'},
            end={'query': past, 'range': 'max'}
        )
        archived_events = [e.id for e in archived_events]

        if archived_events:
            log.info('removing archived events -> %s' % str(archived_events))

            if not dryrun:
                directory.manage_delObjects(archived_events)
        else:
            log.info('no archived events to remove')

        return archived_events

    def remove_past_imported_events(self, directory, dryrun=False):

        catalog = IDirectoryCatalog(directory)
        query = catalog.catalog

        log.info('remove past imported events (> 2 days old)')

        # events are in the past if they have been over for two days
        # (not one, to ensure that they are really over in all timezones)
        past = to_utc(datetime.utcnow() - timedelta(days=2))
        imported_events = query(
            path={'query': directory.getPhysicalPath(), 'depth': 2},
            object_provides=IExternalEvent.__identifier__,
            start={'query': past, 'range': 'max'},
            end={'query': past, 'range': 'max'}
        )

        past_events = []

        for event in imported_events:
            event = event.getObject()

            assert event.start < past
            assert event.end < past

            # recurring events may be in the past with one of
            # their occurrences in the future
            if not has_future_occurrences(event, past):
                past_events.append(event)

        past_events = [p.id for p in past_events]

        if past_events:
            log.info('removing past imported events -> %s' % str(past_events))

            if not dryrun:
                directory.manage_delObjects(past_events)
                pass
        else:
            log.info('no past imported events found')

        return past_events

    def cleanup_directory(self, directory, dryrun=True):

        if dryrun:
            log.info('starting dry run cleanup on %s' %
                     directory.absolute_url())
        else:
            log.info('starting real cleanup on %s' % directory.absolute_url())

        self.remove_stale_previews(directory, dryrun)
        self.archive_past_events(directory, dryrun)
        self.remove_archived_events(directory, dryrun)
        self.remove_past_imported_events(directory, dryrun)

        log.info('finished cleanup on %s' % directory.absolute_url())


cleanup_scheduler = CleanupScheduler()


class CleanupView(grok.View):

    grok.name('cleanup')
    grok.context(IEventsDirectory)
    grok.require('zope2.View')

    def render(self):

        self.request.response.setHeader("Content-type", "text/plain")

        # dryrun must be disabled explicitly using &run=1
        dryrun = not self.request.get('run') == '1'
        force_run = bool(self.request.get('force', False))

        # this maintenance feature may be run unrestricted as it does not
        # leak any information and it's behavior cannot be altered by the
        # user. This allows for easy use via cronjobs.
        execute_under_special_role(
            getSite(), 'Manager',
            cleanup_scheduler.run, self.context, dryrun, force_run
        )

        return u''



# -*- coding: utf-8 -*-

import pytz
import mock

from datetime import datetime, timedelta, date

from zope.publisher.browser import TestRequest

from seantis.dir.events.item import EventsDirectoryItem
from seantis.dir.events.tests import IntegrationTestCase
from seantis.dir.events import recurrence
from seantis.dir.events import dates


class Item(object, EventsDirectoryItem):

    def __init__(self, start, end, recurrence="", timezone='Europe/Zurich',
                 whole_day=False):

        # if this is a whole day, the times are adjusted (this is what
        # plone.app.event would do on the dexterity type)
        if whole_day:
            start = datetime(start.year, start.month, start.day)
            end = datetime(end.year, end.month, end.day, 23, 59, 59)

        # the given date is implicitly of the given timezone, so enforce
        # that and then convert to utc as this is what an EventItem actually
        # stores.
        self.start = dates.to_utc(dates.as_timezone(start, timezone))
        self.end = dates.to_utc(dates.as_timezone(end, timezone))

        self.recurrence = recurrence
        self.timezone = timezone
        self.whole_day = whole_day


class TestRecurrence(IntegrationTestCase):

    def test_occurrences(self):

        item = Item(
            datetime(2012, 1, 1, 10, 0),
            datetime(2012, 1, 1, 12, 0),
            "RRULE:FREQ=DAILY;COUNT=4"
        )

        occurrences = recurrence.occurrences(
            item,
            min_date=datetime(2012, 1, 1, 0, 0, tzinfo=item.tz),
            max_date=datetime(2012, 12, 31, 0, 0, tzinfo=item.tz)
        )

        self.assertEqual(len(occurrences), 4)

        for occurrence in occurrences:
            self.assertEqual(occurrence.recurrence, item.recurrence)

        days = [o.start.day for o in occurrences]
        self.assertEqual([1, 2, 3, 4], days)

        # test occurrences
        item = Item(
            datetime(2012, 1, 1, 10, 0),
            datetime(2012, 1, 1, 12, 0),
            "RRULE:FREQ=DAILY;COUNT=4"
        )

        occurrences = recurrence.occurrences(
            item, min_date=item.start, max_date=item.end
        )

        self.assertEqual(len(occurrences), 1)

        # see if the picking works correctly
        for i in range(1, 5):
            picked = recurrence.pick_occurrence(item, start=date(2012, 1, i))
            self.assertEqual(picked.start.day, i)
            self.assertEqual(picked.start.month, 1)
            self.assertEqual(picked.start.year, 2012)

        picked = recurrence.pick_occurrence(item, start=date(2011, 12, 31))
        self.assertEqual(picked, None)

        picked = recurrence.pick_occurrence(item, start=date(2012, 1, 5))
        self.assertEqual(picked, None)

        # occurrences should return an empty list in these cases
        occurrences = recurrence.occurrences(
            item,
            min_date=datetime(2011, 1, 1, tzinfo=item.tz),
            max_date=datetime(2011, 1, 1, tzinfo=item.tz)
        )

        self.assertEqual(occurrences, [])

        non_recurrant = Item(
            datetime(2012, 1, 1, 10, 0),
            datetime(2012, 1, 1, 12, 0)
        )

        occurrences = recurrence.occurrences(
            non_recurrant,
            min_date=datetime(2011, 1, 1, tzinfo=item.tz),
            max_date=datetime(2011, 1, 1, tzinfo=item.tz)
        )

        self.assertEqual(occurrences, [])

        # non-recurring items are packed into a list
        occurrences = recurrence.occurrences(
            non_recurrant,
            datetime(2012, 1, 1, 10, 0, tzinfo=item.tz),
            datetime(2012, 1, 1, 12, 0, tzinfo=item.tz)
        )

        self.assertEqual(occurrences, [non_recurrant])

    def test_split_days(self):
        # split needed?

        def assert_split(num, daterange):
            start, end = daterange.split(' - ')
            start = datetime.strptime('2012.' + start, '%Y.%d.%m %H:%M')
            end = datetime.strptime('2012.' + end, '%Y.%d.%m %H:%M')
            self.assertEqual(dates.split_days_count(start, end), num)

        assert_split(0, "01.01 08:00 - 01.01 17:00")

        assert_split(0, "01.01 16:00 - 02.01 02:00")
        assert_split(0, "01.01 00:00 - 02.01 08:59")

        assert_split(1, "01.01 16:00 - 02.01 09:00")
        assert_split(1, "01.01 16:00 - 02.01 23:00")
        assert_split(1, "01.01 00:00 - 02.01 23:00")

        assert_split(2, "01.01 00:00 - 03.01 09:00")

        two_days = Item(
            datetime(2012, 1, 1, 10),
            datetime(2012, 1, 2, 20),
            timezone='utc'
        )

        splits = list(recurrence.split_days(two_days))
        self.assertEqual(len(splits), 2)

        self.assertEqual(splits[0].start.hour, 10)
        self.assertEqual(splits[0].end.hour, 23)
        self.assertEqual(splits[0].end.minute, 59)
        self.assertEqual(splits[0].end.second, 59)

        self.assertEqual(splits[1].start.hour, 0)
        self.assertEqual(splits[1].start.minute, 0)
        self.assertEqual(splits[1].start.second, 0)
        self.assertEqual(splits[1].end.hour, 20)

        # one split
        two_days = Item(datetime(2012, 1, 1, 10), datetime(2012, 1, 2, 10))
        splits = list(recurrence.split_days(two_days))

        self.assertEqual(len(splits), 2)

        two_days = Item(datetime(2012, 1, 1, 10), datetime(2012, 1, 2, 9))
        splits = list(recurrence.split_days(two_days))

        self.assertEqual(len(splits), 2)

        # no split
        one_day = Item(datetime(2012, 1, 1, 10), datetime(2012, 1, 2, 8))
        splits = list(recurrence.split_days(one_day))

        self.assertEqual(len(splits), 1)

        # more days
        three_days = Item(datetime(2012, 1, 1, 10), datetime(2012, 1, 3, 20))
        splits = list(recurrence.split_days(three_days))

        self.assertEqual(len(splits), 3)

        self.assertEqual(splits[0].local_start.hour, 10)
        self.assertEqual(splits[0].local_end.hour, 23)
        self.assertEqual(splits[0].local_end.minute, 59)
        self.assertEqual(splits[0].local_end.second, 59)

        self.assertEqual(splits[1].local_start.hour, 0)
        self.assertEqual(splits[1].local_end.hour, 23)
        self.assertEqual(splits[1].local_end.minute, 59)
        self.assertEqual(splits[1].local_end.second, 59)

        self.assertEqual(splits[2].local_start.hour, 0)
        self.assertEqual(splits[2].local_start.minute, 0)
        self.assertEqual(splits[2].local_start.second, 0)
        self.assertEqual(splits[2].local_end.hour, 20)

        # whole_day events
        three_whole_days = Item(
            datetime(2012, 1, 1),
            datetime(2012, 1, 3),
            timezone='utc',
            whole_day=True
        )

        self.assertTrue(
            dates.is_whole_day(three_whole_days.start, three_whole_days.end)
        )

        splits = list(recurrence.split_days(three_whole_days))
        self.assertEqual(len(splits), 3)
        self.assertEqual([1, 2, 3], [s.start.day for s in splits])
        self.assertEqual([1, 2, 3], [s.end.day for s in splits])

    @mock.patch('seantis.dir.events.dates.default_timezone')
    @mock.patch('seantis.dir.events.dates.default_now')
    def test_cornercases(self, default_now, default_timezone):

        default_timezone.return_value = 'Europe/Zurich'

        # create an event that repeats every sunday at 7 in the morning
        # over a period within which daylight savings time changes
        item = Item(
            datetime(2012, 10, 21, 7),
            datetime(2012, 11, 4, 7),
            "RRULE:FREQ=WEEKLY",
            timezone='Europe/Zurich'
        )

        occurrences = recurrence.occurrences(
            item, min_date=item.local_start, max_date=item.local_end
        )

        self.assertEqual(len(occurrences), 3)
        self.assertEqual([o.local_start.hour for o in occurrences], [7] * 3)
        self.assertEqual(
            [o.local_start.weekday() for o in occurrences], [6] * 3
        )

        # create an event that starts at 0:00 in a timezone other than
        # utc and ensure that the resulting human date shows the
        # correct weekday in the given timezone
        # to make it tricky, cross over daylight savings time

        default_now.return_value = datetime(
            2012, 10, 17, tzinfo=pytz.timezone('Europe/Zurich')
        )

        start = datetime(2012, 10, 30)
        item = Item(
            start, start + timedelta(seconds=60 * 60 * 2),
            timezone='Europe/Zurich'
        )

        # behold, the Swiss German language
        request = TestRequest()
        request.locale.dates.calendars['gregorian'].getDayNames = mock.Mock(
            return_value=[
                u'Mäntig', u'Zischtig', u'Mittwuch', u'Dunschtig', u'Fritig',
                u'Samschtig', u'Sunntig'
            ]
        )
        human = item.as_occurrence().human_date(request)

        self.assertTrue(u'Zischtig' in human)

        # while we're at it
        human = item.as_occurrence().human_daterange(request)
        self.assertEqual(human, '00:00 - 02:00')

    def test_has_future_occurrences(self):
        item = Item(
            datetime(2013, 1, 1), datetime(2013, 1, 2), timezone='utc'
        )

        reference = dates.to_utc(datetime(2013, 1, 3))
        self.assertFalse(recurrence.has_future_occurrences(item, reference))

        reference = dates.to_utc(datetime(2013, 1, 2))
        self.assertFalse(recurrence.has_future_occurrences(item, reference))

        reference = dates.to_utc(datetime(2013, 1, 1))
        self.assertTrue(recurrence.has_future_occurrences(item, reference))

        item = Item(
            datetime(2013, 1, 1), datetime(2013, 1, 2), timezone='utc',
                recurrence='RRULE:FREQ=DAILY;UNTIL=20130105T000000Z'
        )

        reference = dates.to_utc(datetime(2013, 1, 4))
        self.assertTrue(recurrence.has_future_occurrences(item, reference))

        reference = dates.to_utc(datetime(2013, 1, 5))
        self.assertTrue(recurrence.has_future_occurrences(item, reference))

        reference = dates.to_utc(datetime(2013, 1, 6))
        self.assertFalse(recurrence.has_future_occurrences(item, reference))

        # the recurrence rule may come without a timezone (missing Z)

        item = Item(
            datetime(2013, 1, 1), datetime(2013, 1, 2), timezone='utc',
                recurrence='RRULE:FREQ=DAILY;UNTIL=20130105T000000'
        )

        reference = dates.to_utc(datetime(2013, 1, 4))
        self.assertTrue(recurrence.has_future_occurrences(item, reference))

        reference = dates.to_utc(datetime(2013, 1, 5))
        self.assertTrue(recurrence.has_future_occurrences(item, reference))

        reference = dates.to_utc(datetime(2013, 1, 6))
        self.assertFalse(recurrence.has_future_occurrences(item, reference))



from datetime import datetime, timedelta

from seantis.dir.events.tests import IntegrationTestCase
from seantis.dir.events import dates


class TestDateRanges(IntegrationTestCase):

    def test_next_weekday(self):
        date = datetime(2012, 9, 18, 0, 0)  # tuesday

        self.assertEqual(date.weekday(), dates.weekdays['TU'])
        self.assertEqual(dates.next_weekday(date, "TU"), date)
        self.assertEqual(
            dates.next_weekday(date, "WE"), date + timedelta(days=1)
        )
        self.assertEqual(
            dates.next_weekday(date, "TH"), date + timedelta(days=2)
        )
        self.assertEqual(
            dates.next_weekday(date, "MO"), date + timedelta(days=6)
        )

    def test_this_weekend(self):
        date = datetime(2012, 9, 18, 0, 0)  # tuesday
        start, end = dates.this_weekend(date)

        self.assertEqual((start.year, start.month, start.day), (2012, 9, 21))
        self.assertEqual(start.weekday(), dates.weekdays["FR"])
        self.assertEqual(start.hour, 16)

        self.assertEqual((end.year, end.month, end.day), (2012, 9, 23))
        self.assertEqual(end.weekday(), dates.weekdays["SU"])
        self.assertEqual(end.hour, 23)
        self.assertEqual(end.minute, 59)
        self.assertEqual(end.second, 59)

        date = datetime(2012, 9, 23, 0, 0)  # sunday
        start, end = dates.this_weekend(date)

        self.assertEqual((start.year, start.month, start.day), (2012, 9, 21))
        self.assertEqual((end.year, end.month, end.day), (2012, 9, 23))

        old_start, old_end = start, end

        start, end = dates.this_weekend(start)
        self.assertEqual(start, old_start)
        self.assertEqual(end, old_end)

        start, end = dates.this_weekend(end)
        self.assertEqual(start, old_start)
        self.assertEqual(end, old_end)

    def test_morning(self):

        dateranges = dates.DateRanges()
        dateranges.now = datetime.today()
        overlaps = lambda: dateranges.overlaps('today', start, end)

        start = datetime.today().replace(hour=1, minute=0)
        end = datetime.today().replace(hour=4, minute=0)
        self.assertTrue(overlaps())

        start = datetime.today().replace(hour=1, minute=0)
        end = datetime.today().replace(hour=3, minute=0)
        self.assertTrue(overlaps())

        start = datetime.today().replace(hour=1, minute=0)
        end = datetime.today().replace(hour=2, minute=59)
        self.assertFalse(overlaps())

    def test_ranges(self):

        start = datetime(2012, 1, 1, 22, 0)  # sunday
        end = datetime(2012, 1, 2, 03, 0)
        dateranges = dates.DateRanges()
        overlaps = lambda method: dateranges.overlaps(method, start, end)

        dateranges.now = datetime(2011, 1, 1, 0, 0)
        self.assertFalse(overlaps('today'))
        self.assertFalse(overlaps('this_year'))
        self.assertTrue(overlaps('next_year'))
        self.assertTrue(overlaps('this_and_next_year'))

        dateranges.now = datetime(2012, 1, 1, 12, 0)
        self.assertTrue(overlaps('today'))
        self.assertTrue(overlaps('this_year'))
        self.assertTrue(overlaps('this_and_next_year'))

        dateranges.now = datetime(2011, 12, 31, 12, 0)  # saturday
        self.assertFalse(overlaps('today'))
        self.assertTrue(overlaps('tomorrow'))
        self.assertTrue(overlaps('this_week'))

        dateranges.now = datetime(2011, 12, 24, 12, 0)  # sat. one week before
        self.assertTrue(overlaps('this_week'))
        self.assertTrue(overlaps('next_week'))  # spillover

        dateranges.now = datetime(2011, 12, 23, 12, 0)  # fri. one week before
        self.assertFalse(overlaps('this_week'))
        self.assertTrue(overlaps('next_week'))

        dateranges.now = datetime(2012, 1, 2, 12, 0)
        self.assertTrue(overlaps('this_week'))

        dateranges.now = datetime(2011, 12, 29, 0, 0)  # thursday
        self.assertTrue(overlaps('this_week'))
        self.assertTrue(overlaps('this_weekend'))
        self.assertTrue(overlaps('next_week'))  # spillover
        self.assertFalse(overlaps('next_weekend'))

        dateranges.now = datetime(2011, 12, 22, 0, 0)  # one week earlier
        self.assertFalse(overlaps('this_week'))
        self.assertFalse(overlaps('this_weekend'))
        self.assertTrue(overlaps('next_week'))
        self.assertTrue(overlaps('next_weekend'))

        dateranges.now = datetime(2011, 12, 12, 0, 0)
        self.assertFalse(overlaps('this_month'))
        self.assertTrue(overlaps('next_month'))

        dateranges.now = datetime(2012, 1, 1, 0, 0)
        self.assertTrue(overlaps('this_month'))
        self.assertFalse(overlaps('next_month'))



import pytz
import urllib

from datetime import datetime

from five import grok

from Products.CMFCore.utils import getToolByName
from Products.statusmessages.interfaces import IStatusMessage

from plone.memoize import view
from plone.app.event.ical.exporter import construct_icalendar
from plone.event.interfaces import IICalendarEventComponent
from plone.app.event.ical import ICalendarEventComponent
from OFS.interfaces import IObjectClonedEvent
from zExceptions import NotFound

from seantis.dir.base import item
from seantis.dir.base import core
from seantis.dir.base.interfaces import (
    IFieldMapExtender,
    IDirectoryCategorized
)

from seantis.dir.events import _
from seantis.dir.events import dates
from seantis.dir.events import recurrence
from seantis.dir.events import utils
from seantis.dir.events.token import verify_token
from seantis.dir.events.interfaces import (
    IEventsDirectory, IEventsDirectoryItem, IExternalEvent
)

from AccessControl import getSecurityManager
from Products.CMFCore import permissions


# when an event is cloned put it in the submitted state
@grok.subscribe(IEventsDirectoryItem, IObjectClonedEvent)
def onClonedEvent(item, event):
    item.submit()


class EventsDirectoryItem(item.DirectoryItem):

    actions_order = ('submit', 'publish', 'deny', 'archive', 'hide')

    # there's currently an issue with seantis.dir.events, plone.app.event and
    # plone.app.imaging. this mock function servers as a work-around.
    # see https://github.com/plone/plone.app.event/issues/78
    def Schema(self):

        class MockSchema(object):
            def get(self, field):
                return None

        return MockSchema()

    def get_parent(self):
        return self.aq_inner.aq_parent

    @property
    def tz(self):
        return pytz.timezone(self.timezone)

    @property
    def local_start(self):
        return self.tz.normalize(self.start)

    @property
    def local_end(self):
        return self.tz.normalize(self.end)

    def as_occurrence(self):
        return recurrence.Occurrence(self, self.start, self.end)

    @property
    def state(self):
        """ Return the workflow state. """
        return utils.workflow_tool().getInfoFor(self, 'review_state')

    @property
    def review_state(self):
        return self.state

    def action_url(self, action):
        baseurl = self.absolute_url() + '/do-action?action=%s'
        return baseurl % action['id']

    def list_actions(self):
        sortkey = lambda a: self.actions_order.index(a['id'])

        workflowTool = getToolByName(self, "portal_workflow")
        return sorted(workflowTool.listActions(object=self), key=sortkey)

    def do_action(self, action):
        """ Execute the given action. """
        workflowTool = getToolByName(self, "portal_workflow")
        workflowTool.doActionFor(self, action)

    def allow_action(self, action):
        return self.get_parent().allow_action(action, self)

    def submit(self):
        self.do_action("submit")

    def deny(self):
        self.do_action("deny")

    def publish(self):
        self.do_action("publish")

    def archive(self):
        self.do_action("archive")

    def hide(self):
        self.do_action("hide")

    def reindex(self):
        utils.get_catalog(self.get_parent()).reindex()

    def attachment_filename(self, attachment):
        filename = getattr(self, attachment).filename

        if not filename:
            number = attachment[-1]
            return _(u'Attachment ${number}', mapping=dict(number=number))

        if len(filename) > 100:
            return filename[:100] + '...'
        else:
            return filename

    def eventtags(self):
        """ Return a list of tuples containing the event tag value
        (category value) in position 0 and the link to the related
        filter in position 1.

        The results are sorted by value (position 0).
        """

        categorized = IDirectoryCategorized(self)

        categories = dict()
        categories['cat1'] = categorized.keywords(categories=('cat1', ))
        categories['cat2'] = categorized.keywords(categories=('cat2', ))

        baseurl = self.get_parent().absolute_url() + '?filter=true&%s=%s'

        tags = list()

        for key in sorted(categories):
            for tag in sorted(categories[key]):
                if not tag:
                    continue

                tags.append((
                    tag.strip().replace(' ', '&nbsp;'),
                    baseurl % (key, urllib.quote(tag.encode('utf-8')))
                ))

        return tags

    @property
    def allow_edit(self):
        return not IExternalEvent.providedBy(self)


class DoActionView(grok.View):
    """ Pretty much like modify_content_status, but with
    better messages and redirection to the directory. """

    grok.name('do-action')
    grok.context(IEventsDirectoryItem)
    grok.require('cmf.ReviewPortalContent')

    messages = {
        'publish': _(u'Event was published'),
        'archive': _(u'Event was archived'),
        'deny': _(u'Publication of event was denied'),
        'hide': _(u'Event was hidden')
    }

    def render(self):

        action = self.request.get('action')
        assert action in self.messages

        IStatusMessage(self.request).add(self.messages[action], "info")
        self.context.do_action(action)
        self.request.response.redirect(
            self.context.get_parent().absolute_url()
        )

        return ""


class View(core.View):
    """Default view of a seantis.dir.events item."""
    grok.context(IEventsDirectoryItem)
    grok.require('zope2.View')

    # don't show the search viewlet the item's view
    hide_search_viewlet = True

    template = None
    _template = grok.PageTemplateFile('templates/item.pt')

    def update(self):
        verify_token(self.context, self.request)

        super(View, self).update()

    @property
    def is_ical_export(self):
        return self.request.get('type') == 'ical'

    @property
    def is_valid_date(self):
        return self.occurrence is not None

    def render(self):
        if not self.is_valid_date:
            raise NotFound()

        if not self.is_ical_export:
            return self._template.render(self)
        else:
            if self.date:
                calendar = construct_icalendar(
                    self.context, [self.occurrence]
                )
                for component in calendar.subcomponents:
                    if 'RRULE' in component:
                        del component['RRULE']
            else:
                calendar = construct_icalendar(
                    self.context, [self.context]
                )

            return utils.render_ical_response(self.request, self.context, calendar)

    @utils.webcal
    def ical_url(self, only_this):
        url = self.context.absolute_url() + '?type=ical'
        if only_this and self.request.get('date'):
            url += '&date=' + self.request.get('date')
        return url

    @property
    def is_recurring(self):
        return self.context.recurrence and True or False

    def recurrence_url(self, event):
        baseurl = self.context.get_parent().absolute_url()
        baseurl += '?range=this_and_next_year&search=true&searchtext=%s'
        return baseurl % urllib.quote(event.short_description.encode('utf-8'))

    @property
    def date(self):
        date = self.request.get('date')
        if not date:
            return None

        try:
            return datetime.strptime(date, '%Y-%m-%d')
        except:
            return None

    @property
    def human_date(self):
        return self.occurrence.human_date(self.request)

    @property
    def human_daterange(self):
        return self.occurrence.human_daterange(self.request)

    @property
    @view.memoize
    def occurrence(self):
        date = self.date
        if date and self.is_recurring:
            return recurrence.pick_occurrence(self.context, self.date)
        else:
            return self.context.as_occurrence()

    @property
    @view.memoize
    def occurrences(self):
        min_date, max_date = dates.eventrange()

        return recurrence.occurrences(self.context, min_date, max_date)

    @property
    def show_submitter(self):
        # if the information not present, do not show
        if not all((self.context.submitter, self.context.submitter_email)):
            return False

        # someone who may open the edit view can see the submitter there,
        # so we can display it on the detail view in this case
        return getSecurityManager().checkPermission(
            permissions.ModifyPortalContent, self.context
        )


class ICalendarEventItemComponent(ICalendarEventComponent, grok.Adapter):
    """ Adds custom information to the default ical implementation. """

    grok.implements(IICalendarEventComponent)
    grok.context(IEventsDirectoryItem)

    def get_coordinates(self):

        coordinates = self.context.get_coordinates()

        if not len(coordinates) == 2:
            return None

        if not (coordinates[0] or '').upper() == 'POINT':
            return None

        return coordinates[1]

    def to_ical(self):
        ical = ICalendarEventComponent.to_ical(self)

        coordinates = self.get_coordinates()
        if coordinates:
            # ical actually needs a tuple here, not a list which I argue is
            # a bug: https://github.com/collective/icalendar/issues/83
            ical.add('geo', tuple(coordinates))

        # plone.app.event.ical does the following, but when I tried
        # to use their interfaces for contact and location I got really
        # weird problems that made me give up at some point out of fed-up-ness
        e = self.context

        if e.locality:
            ical.add('location', e.locality)

        contact = [
            e.contact_name, e.contact_phone, e.contact_email, e.event_url
        ]
        contact = filter(lambda c: c, contact)

        if contact:
            ical.add('contact', u', '.join(contact))

        return ical


class ExtendedDirectoryItemFieldMap(grok.Adapter):
    """Adapter extending the import/export fieldmap of
    seantis.dir.events.item.

    """
    grok.context(IEventsDirectory)
    grok.provides(IFieldMapExtender)

    def __init__(self, context):
        self.context = context

    def extend_import(self, itemmap):
        itemmap.typename = 'seantis.dir.events.item'
        itemmap.interface = IEventsDirectoryItem

        extended = [
            'start', 'end', 'timezone', 'whole_day', 'recurrence',
            'short_description', 'long_description', 'locality', 'street',
            'housenumber', 'zipcode', 'town', 'location_url', 'event_url',
            'organizer', 'contact_name', 'contact_email', 'contact_phone',
            'prices', 'registration', 'submitter', 'submitter_email'
        ]

        boolwrap = lambda v: v and '1' or ''
        boolunwrap = lambda v: v == '1'

        itemmap.bind_wrapper("whole_day", boolwrap)
        itemmap.bind_unwrapper("whole_day", boolunwrap)

        datewrap = lambda v: v.strftime('%Y.%m.%d %H-%M')
        dateunwrap = lambda v: datetime.strptime(v, '%Y.%m.%d %H-%M')

        itemmap.bind_wrapper("start", datewrap)
        itemmap.bind_unwrapper("start", dateunwrap)

        itemmap.bind_wrapper("end", datewrap)
        itemmap.bind_unwrapper("end", dateunwrap)

        itemmap.add_fields(extended, len(itemmap))



from logging import getLogger
log = getLogger('seantis.dir.events')

import re
import threading

from five import grok

from ZServer.ClockServer import ClockServer

from seantis.dir.events.interfaces import IResourceViewedEvent


_clockservers = dict()  # list of registered Zope Clockservers
_lock = threading.Lock()


# The primary hook to setup maintenance clockservers is the reservation view
# event. It's invoked way too often, but the invocation is fast and it is
# guaranteed to be run on a plone site with seantis.events installed,
# setup and in use. Other events like zope startup and traversal events are
# not safe enough to use if one has to rely on a site being setup.
@grok.subscribe(IResourceViewedEvent)
def on_resource_viewed(event):
    path = '/'.join(event.context.getPhysicalPath())
    register(path + '/fetch?run=1', 15 * 60)
    register(path + '/cleanup?run=1', 60 * 60)


def clear_clockservers():
    """ Clears the clockservers and connections for testing. """

    with _lock:
        for cs in _clockservers.values():
            cs.close()
        _clockservers.clear()


def register(method, period):
    """ Registers the given method with a clockserver.

    Note that due to it's implementation there's no guarantee that the method
    will be called on time every time. The clockserver checks if a call is due
    every time a request comes in, or every 30 seconds when the asyncore.pollss
    method reaches it's timeout (see Lifetime/__init__.py and
    Python Stdlib/asyncore.py).

    """

    if method not in _clockservers:
        with _lock:
            _clockservers[method] = ClockServer(
                method, period, host='localhost', logger=ClockLogger(method)
            )

    return _clockservers[method]


logexpr = re.compile(r'GET [^\s]+ HTTP/[^\s]+ ([0-9]+)')


class ClockLogger(object):

    """ Logs the clock runs by evaluating the log strings. Looks for http
    return codes to do so.

    """

    def __init__(self, method):
        self.method = method

    def return_code(self, msg):
        groups = re.findall(logexpr, msg)
        return groups and int(groups[0]) or None

    def log(self, msg):
        code = self.return_code(msg)

        if not code:
            log.error("ClockServer for %s returned nothing" % self.method)
        elif code == 200:
            pass
        else:
            log.warn("ClockServer for %s returned %i" % (self.method, code))
